"""Linear operator for the Fisher/GGN's Kronecker-factored approximation.

Kronecker-Factored Approximate Curvature (KFAC) was originally introduced for MLPs in

- Martens, J., & Grosse, R. (2015). Optimizing neural networks with Kronecker-factored
  approximate curvature. International Conference on Machine Learning (ICML),

extended to CNNs in

- Grosse, R., & Martens, J. (2016). A kronecker-factored approximate Fisher matrix for
  convolution layers. International Conference on Machine Learning (ICML),

and generalized to all linear layers with weight sharing in

- Eschenhagen, R., Immer, A., Turner, R. E., Schneider, F., Hennig, P. (2023).
  Kronecker-Factored Approximate Curvature for Modern Neural Network Architectures (NeurIPS).
"""

from __future__ import annotations

from collections.abc import MutableMapping
from enum import Enum, EnumMeta
from functools import partial
from math import sqrt
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

from einops import einsum, rearrange, reduce
from numpy import ndarray
from torch import Generator, Tensor, cat, device, eye, randn, stack
from torch.autograd import grad
from torch.nn import (
    BCEWithLogitsLoss,
    Conv2d,
    CrossEntropyLoss,
    Linear,
    Module,
    MSELoss,
    Parameter,
)
from torch.utils.hooks import RemovableHandle

from curvlinops._base import _LinearOperator
from curvlinops.kfac_utils import (
    extract_averaged_patches,
    extract_patches,
    loss_hessian_matrix_sqrt,
)

# Type for a matrix/vector that can be represented as a list of tensors with the same
# shape as the parameters, or a single matrix/vector of shape `[D, D]`/`[D]` where `D`
# is the number of parameters.
ParameterMatrixType = TypeVar("ParameterMatrixType", Tensor, List[Tensor])


class MetaEnum(EnumMeta):
    """Metaclass for the Enum class for desired behavior of the `in` operator."""

    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class FisherType(str, Enum, metaclass=MetaEnum):
    """Enum for the Fisher type."""

    TYPE2 = "type-2"
    MC = "mc"
    EMPIRICAL = "empirical"
    FORWARD_ONLY = "forward-only"


class KFACType(str, Enum, metaclass=MetaEnum):
    """Enum for the KFAC approximation type."""

    EXPAND = "expand"
    REDUCE = "reduce"


class KFACLinearOperator(_LinearOperator):
    r"""Linear operator to multiply with the Fisher/GGN's KFAC approximation.

    KFAC approximates the per-layer Fisher/GGN with a Kronecker product:
    Consider a weight matrix :math:`\mathbf{W}` and a bias vector :math:`\mathbf{b}`
    in a single layer. The layer's Fisher :math:`\mathbf{F}(\mathbf{\theta})` for

    .. math::
        \mathbf{\theta}
        =
        \begin{pmatrix}
        \mathrm{vec}(\mathbf{W}) \\ \mathbf{b}
        \end{pmatrix}

    where :math:`\mathrm{vec}` denotes column-stacking is approximated as

    .. math::
        \mathbf{F}(\mathbf{\theta})
        \approx
        \mathbf{A}_{(\text{KFAC})} \otimes \mathbf{B}_{(\text{KFAC})}

    (see :class:`curvlinops.FisherMCLinearOperator` for the Fisher's definition).
    Loosely speaking, the first Kronecker factor is the un-centered covariance of the
    inputs to a layer. The second Kronecker factor is the un-centered covariance of
    'would-be' gradients w.r.t. the layer's output. Those 'would-be' gradients result
    from sampling labels from the model's distribution and computing their gradients.

    Kronecker-Factored Approximate Curvature (KFAC) was originally introduced for MLPs in

    - Martens, J., & Grosse, R. (2015). Optimizing neural networks with Kronecker-factored
    approximate curvature. International Conference on Machine Learning (ICML),

    extended to CNNs in

    - Grosse, R., & Martens, J. (2016). A kronecker-factored approximate Fisher matrix for
    convolution layers. International Conference on Machine Learning (ICML),

    and generalized to all linear layers with weight sharing in

    - Eschenhagen, R., Immer, A., Turner, R. E., Schneider, F., Hennig, P. (2023).
    Kronecker-Factored Approximate Curvature for Modern Neural Network Architectures (NeurIPS).

    Attributes:
        _SUPPORTED_LOSSES: Tuple of supported loss functions.
        _SUPPORTED_MODULES: Tuple of supported layers.
    """

    _SUPPORTED_LOSSES = (MSELoss, CrossEntropyLoss, BCEWithLogitsLoss)
    _SUPPORTED_MODULES = (Linear, Conv2d)
    _SUPPORTED_FISHER_TYPE: FisherType = FisherType
    _SUPPORTED_KFAC_APPROX: KFACType = KFACType

    def __init__(
        self,
        model_func: Module,
        loss_func: Union[MSELoss, CrossEntropyLoss, BCEWithLogitsLoss],
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        progressbar: bool = False,
        check_deterministic: bool = True,
        shape: Union[Tuple[int, int], None] = None,
        seed: int = 2147483647,
        fisher_type: str = FisherType.MC,
        mc_samples: int = 1,
        kfac_approx: str = KFACType.EXPAND,
        num_per_example_loss_terms: Optional[int] = None,
        separate_weight_and_bias: bool = True,
        num_data: Optional[int] = None,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ):
        """Kronecker-factored approximate curvature (KFAC) proxy of the Fisher/GGN.

        Warning:
            If the model's parameters change, e.g. during training, you need to
            create a fresh instance of this object. This is because, for performance
            reasons, the Kronecker factors are computed once and cached during the
            first matrix-vector product. They will thus become outdated if the model
            changes.

        Warning:
            This is an early proto-type with limitations:
                - Only Linear and Conv2d modules are supported.
                - If ``deterministic_checks`` is turned on (as is by default), this
                  will compute the KFAC matrices on CPU, even if all passed arguments
                  live on the GPU.

        Args:
            model_func: The neural network. Must consist of modules.
            loss_func: The loss function.
            params: The parameters defining the Fisher/GGN that will be approximated
                through KFAC.
            data: A data loader containing the data of the Fisher/GGN.
            progressbar: Whether to show a progress bar when computing the Kronecker
                factors. Defaults to ``False``.
            check_deterministic: Whether to check that the linear operator is
                deterministic. Defaults to ``True``.
            shape: The shape of the linear operator. If ``None``, it will be inferred
                from the parameters. Defaults to ``None``.
            seed: The seed for the random number generator used to draw labels
                from the model's predictive distribution. Defaults to ``2147483647``.
            fisher_type: The type of Fisher/GGN to approximate.
                If ``FisherType.TYPE2``, the exact Hessian of the loss w.r.t. the model
                outputs is used. This requires as many backward passes as the output
                dimension, i.e. the number of classes for classification. This is
                sometimes also called type-2 Fisher. If ``FisherType.MC``, the
                expectation is approximated by sampling ``mc_samples`` labels from the
                model's predictive distribution. If ``FisherType.EMPIRICAL``, the
                empirical gradients are used which corresponds to the uncentered
                gradient covariance, or the empirical Fisher.
                If ``FisherType.FORWARD_ONLY``, the gradient covariances will be
                identity matrices, see the FOOF method in
                `Benzing, 2022 <https://arxiv.org/abs/2201.12250>`_ or ISAAC in
                `Petersen et al., 2023 <https://arxiv.org/abs/2305.00604>`_.
                Defaults to ``FisherType.MC``.
            mc_samples: The number of Monte-Carlo samples to use per data point.
                Has to be set to ``1`` when ``fisher_type != FisherType.MC``.
                Defaults to ``1``.
            kfac_approx: A string specifying the KFAC approximation that should
                be used for linear weight-sharing layers, e.g. ``Conv2d`` modules
                or ``Linear`` modules that process matrix- or higher-dimensional
                features.
                Possible values are ``KFACType.EXPAND`` and ``KFACType.REDUCE``.
                See `Eschenhagen et al., 2023 <https://arxiv.org/abs/2311.00636>`_
                for an explanation of the two approximations.
                Defaults to ``KFACType.EXPAND``.
            num_per_example_loss_terms: Number of per-example loss terms, e.g., the
                number of tokens in a sequence. The model outputs will have
                ``num_data * num_per_example_loss_terms * C`` entries, where ``C`` is
                the dimension of the random variable we define the likelihood over --
                for the ``CrossEntropyLoss`` it will be the number of classes, for the
                ``MSELoss`` and ``BCEWithLogitsLoss`` it will be the size of the last
                dimension of the the model outputs/targets (our convention here).
                If ``None``, ``num_per_example_loss_terms`` is inferred from the data at
                the cost of one traversal through the data loader. It is expected to be
                the same for all examples. Defaults to ``None``.
            separate_weight_and_bias: Whether to treat weights and biases separately.
                Defaults to ``True``.
            num_data: Number of data points. If ``None``, it is inferred from the data
                at the cost of one traversal through the data loader.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.

        Raises:
            RuntimeError: If the check for deterministic behavior fails.
            ValueError: If the loss function is not supported.
            ValueError: If ``fisher_type != FisherType.MC`` and ``mc_samples != 1``.
            ValueError: If ``X`` is not a tensor and ``batch_size_fn`` is not specified.
        """
        if not isinstance(loss_func, self._SUPPORTED_LOSSES):
            raise ValueError(
                f"Invalid loss: {loss_func}. Supported: {self._SUPPORTED_LOSSES}."
            )
        if fisher_type not in self._SUPPORTED_FISHER_TYPE:
            raise ValueError(
                f"Invalid fisher_type: {fisher_type}. "
                f"Supported: {self._SUPPORTED_FISHER_TYPE}."
            )
        if fisher_type != FisherType.MC and mc_samples != 1:
            raise ValueError(
                f"Invalid mc_samples: {mc_samples}. "
                "Only mc_samples=1 is supported for `fisher_type != FisherType.MC`."
            )
        if kfac_approx not in self._SUPPORTED_KFAC_APPROX:
            raise ValueError(
                f"Invalid kfac_approx: {kfac_approx}. "
                f"Supported: {self._SUPPORTED_KFAC_APPROX}."
            )

        self._seed = seed
        self._generator: Union[None, Generator] = None
        self._separate_weight_and_bias = separate_weight_and_bias
        self._fisher_type = fisher_type
        self._mc_samples = mc_samples
        self._kfac_approx = kfac_approx
        self._input_covariances: Dict[str, Tensor] = {}
        self._gradient_covariances: Dict[str, Tensor] = {}
        self._mapping = self.compute_parameter_mapping(params, model_func)

        # Properties of the full matrix KFAC approximation are initialized to `None`
        self._reset_matrix_properties()

        super().__init__(
            model_func,
            loss_func,
            params,
            data,
            progressbar=progressbar,
            check_deterministic=False,
            shape=shape,
            num_data=num_data,
            batch_size_fn=batch_size_fn,
        )

        self._set_num_per_example_loss_terms(num_per_example_loss_terms)

        if check_deterministic:
            old_device = self._device
            self.to_device(device("cpu"))
            try:
                self._check_deterministic()
            except RuntimeError as e:
                raise e
            finally:
                self.to_device(old_device)

    def _set_num_per_example_loss_terms(
        self, num_per_example_loss_terms: Optional[int]
    ):
        """Set the number of per-example loss terms.

        Args:
            num_per_example_loss_terms: Number of per-example loss terms. If ``None``,
                it is inferred from the data at the cost of one traversal through the
                data loader.

        Raises:
            ValueError: If the number of loss terms is not divisible by the number of
                data points.
        """
        if num_per_example_loss_terms is None:
            # Determine the number of per-example loss terms
            num_loss_terms = sum(
                (
                    y.numel()
                    if isinstance(self._loss_func, CrossEntropyLoss)
                    else y.shape[:-1].numel()
                )
                for (_, y) in self._loop_over_data(desc="_num_per_example_loss_terms")
            )
            if num_loss_terms % self._N_data != 0:
                raise ValueError(
                    "The number of loss terms must be divisible by the number of data "
                    f"points; num_loss_terms={num_loss_terms}, N_data={self._N_data}."
                )
            self._num_per_example_loss_terms = num_loss_terms // self._N_data
        else:
            self._num_per_example_loss_terms = num_per_example_loss_terms

    def _reset_matrix_properties(self):
        """Reset matrix properties."""
        self._trace = None
        self._det = None
        self._logdet = None
        self._frobenius_norm = None

    def to_device(self, device: device):
        """Load the linear operator to another device.

        Args:
            device: The device to which the linear operator should be moved.
        """
        super().to_device(device)
        for key in self._input_covariances.keys():
            self._input_covariances[key] = self._input_covariances[key].to(device)
        for key in self._gradient_covariances.keys():
            self._gradient_covariances[key] = self._gradient_covariances[key].to(device)

    def _torch_preprocess(self, M: Tensor) -> List[Tensor]:
        """Convert torch tensor to torch parameter list format.

        Args:
            M: Matrix for multiplication. Has shape ``[D, K]`` where ``D`` is the
                number of parameters, and ``K`` is the number of columns.

        Returns:
            Matrix in list format. Each entry has the same shape as a parameter with
            an additional leading dimension of size ``K`` for the columns, i.e.
            ``[(K,) + p1.shape, (K,) + p2.shape, ...]``.
        """
        num_vectors = M.shape[1]
        # split parameter blocks
        dims = [p.numel() for p in self._params]
        result = M.split(dims)
        # column-index first + unflatten parameter dimension
        shapes = [(num_vectors,) + p.shape for p in self._params]
        return [res.T.reshape(shape) for res, shape in zip(result, shapes)]

    def _check_input_type_and_preprocess(
        self, M_torch: ParameterMatrixType
    ) -> Tuple[bool, List[Tensor]]:
        """Check input type and maybe preprocess to list format.

        Check whether the input is a tensor or a list of tensors. If it is a tensor,
        preprocess to list format.

        Args:
            M_torch: Input to check.

        Returns:
            ``True`` if the input is a tensor, ``False`` if it is a list of tensors.

        Raises:
            ValueError: If the input is a list of tensors that have a different number
                of columns.
            ValueError: If the input is a list of tensors that have incompatible shapes
                with the parameters.
            ValueError: If the input is a tensor and has the wrong shape.
            ValueError: If the input is a tensor and its shape is incompatible with the
                KFAC approximation's shape.
        """
        if isinstance(M_torch, list):
            return_tensor = False
            if len(M_torch) != len(self._params):
                raise ValueError(
                    "Number of input tensors must match the number of parameter tensors."
                )
            column_values = {M.shape[0] for M in M_torch}
            if len(column_values) != 1:
                raise ValueError(
                    "Number of columns must be equal for all tensors. "
                    f"Got {column_values}."
                )
            K = column_values.pop()
            for M, p in zip(M_torch, self._params):
                if M.shape != (K,) + p.shape:
                    raise ValueError(
                        "All input tensors must have (K, ) + p.shape. "
                        f"Got {M.shape}, but expected {(K,) + p.shape}."
                    )
        else:
            return_tensor = True
            if M_torch.ndim != 2:
                raise ValueError(f"expected 2-d tensor, not {M_torch.ndim}-d")
            if M_torch.shape[0] != self.shape[1]:
                raise ValueError(f"dimension mismatch: {self.shape}, {M_torch.shape}")
            M_torch = self._torch_preprocess(M_torch)
        return return_tensor, M_torch

    def torch_matmat(self, M_torch: ParameterMatrixType) -> ParameterMatrixType:
        """Apply KFAC to a matrix (multiple vectors) in PyTorch.

        This allows for matrix-matrix products with the KFAC approximation in PyTorch
        without converting tensors to numpy arrays, which avoids unnecessary
        device transfers when working with GPUs and flattening/concatenating.

        Args:
            M_torch: Matrix for multiplication. If list of tensors, each entry has the
                same shape as a parameter with an additional leading dimension of size
                ``K`` for the columns, i.e. ``[(K,) + p1.shape), (K,) + p2.shape, ...]``.
                If tensor, has shape ``[D, K]`` with some ``K``.

        Returns:
            Matrix-multiplication result ``KFAC @ M``. Return type is the same as the
            type of the input. If list of tensors, each entry has the same shape as a
            parameter with an additional leading dimension of size ``K`` for the columns,
            i.e. ``[(K,) + p1.shape, (K,) + p2.shape, ...]``. If tensor, has shape
            ``[D, K]`` with some ``K``.
        """
        return_tensor, M_torch = self._check_input_type_and_preprocess(M_torch)
        if not self._input_covariances and not self._gradient_covariances:
            self._compute_kfac()

        for mod_name, param_pos in self._mapping.items():
            # cache the weight shape to ensure correct shapes are returned
            if "weight" in param_pos:
                weight_shape = M_torch[param_pos["weight"]].shape

            # bias and weights are treated jointly
            if (
                not self._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                w_pos, b_pos = param_pos["weight"], param_pos["bias"]
                # v denotes the free dimension for treating multiple vectors in parallel
                M_w = rearrange(M_torch[w_pos], "v c_out ... -> v c_out (...)")
                M_joint = cat([M_w, M_torch[b_pos].unsqueeze(-1)], dim=2)
                aaT = self._input_covariances[mod_name]
                ggT = self._gradient_covariances[mod_name]
                M_joint = einsum(ggT, M_joint, aaT, "i j,v j k,k l -> v i l")

                w_cols = M_w.shape[2]
                M_torch[w_pos], M_torch[b_pos] = M_joint.split([w_cols, 1], dim=2)

            # for weights we need to multiply from the right with aaT
            # for weights and biases we need to multiply from the left with ggT
            else:
                for p_name, pos in param_pos.items():
                    if p_name == "weight":
                        M_w = rearrange(M_torch[pos], "v c_out ... -> v c_out (...)")
                        M_torch[pos] = einsum(
                            M_w,
                            self._input_covariances[mod_name],
                            "v c_out j,j k -> v c_out k",
                        )

                    M_torch[pos] = einsum(
                        self._gradient_covariances[mod_name],
                        M_torch[pos],
                        "j k,v k ... -> v j ...",
                    )

            # restore original shapes
            if "weight" in param_pos:
                M_torch[param_pos["weight"]] = M_torch[param_pos["weight"]].view(
                    weight_shape
                )

        if return_tensor:
            M_torch = cat([rearrange(M, "k ... -> (...) k") for M in M_torch])

        return M_torch

    def torch_matvec(self, v_torch: ParameterMatrixType) -> ParameterMatrixType:
        """Apply KFAC to a vector in PyTorch.

        This allows for matrix-vector products with the KFAC approximation in PyTorch
        without converting tensors to numpy arrays, which avoids unnecessary
        device transfers when working with GPUs and flattening/concatenating.

        Args:
            v_torch: Vector for multiplication. If list of tensors, each entry has the
                same shape as a parameter, i.e. ``[p1.shape, p2.shape, ...]``.
                If tensor, has shape ``[D]``.

        Returns:
            Matrix-multiplication result ``KFAC @ v``. Return type is the same as the
            type of the input. If list of tensors, each entry has the same shape as a
            parameter, i.e. ``[p1.shape, p2.shape, ...]``. If tensor, has shape ``[D]``.

        Raises:
            ValueError: If the input tensor has the wrong data type.
        """
        if isinstance(v_torch, list):
            v_torch = [v_torch_i.unsqueeze(0) for v_torch_i in v_torch]
            result = self.torch_matmat(v_torch)
            return [res.squeeze(0) for res in result]
        elif isinstance(v_torch, Tensor):
            return self.torch_matmat(v_torch.unsqueeze(-1)).squeeze(-1)
        else:
            raise ValueError(
                f"Invalid input type: {type(v_torch)}. Expected list of tensors or tensor."
            )

    def _matmat(self, M: ndarray) -> ndarray:
        """Apply KFAC to a matrix (multiple vectors).

        Args:
            M: Matrix for multiplication. Has shape ``[D, K]`` with some ``K``.

        Returns:
            Matrix-multiplication result ``KFAC @ M``. Has shape ``[D, K]``.
        """
        M_torch = super()._preprocess(M)
        M_torch = self.torch_matmat(M_torch)
        return self._postprocess(M_torch)

    def _adjoint(self) -> KFACLinearOperator:
        """Return the linear operator representing the adjoint.

        The KFAC approximation is real symmetric, and hence self-adjoint.

        Returns:
            Self.
        """
        return self

    def _compute_kfac(self):
        """Compute and cache KFAC's Kronecker factors for future ``matmat``s."""
        self._reset_matrix_properties()

        # install forward and backward hooks
        hook_handles: List[RemovableHandle] = []

        for mod_name, param_pos in self._mapping.items():
            module = self._model_func.get_submodule(mod_name)

            # input covariance only required for weights
            if "weight" in param_pos.keys():
                hook_handles.append(
                    module.register_forward_pre_hook(
                        partial(
                            self._hook_accumulate_input_covariance, module_name=mod_name
                        )
                    )
                )

            # gradient covariance required for weights and biases
            hook_handles.append(
                module.register_forward_hook(
                    partial(
                        self._register_tensor_hook_on_output_to_accumulate_gradient_covariance,
                        module_name=mod_name,
                    )
                )
            )

        # loop over data set, computing the Kronecker factors
        if self._generator is None or self._generator.device != self._device:
            self._generator = Generator(device=self._device)
        self._generator.manual_seed(self._seed)

        for X, y in self._loop_over_data(desc="KFAC matrices"):
            output = self._model_func(X)
            self._compute_loss_and_backward(output, y)

        # clean up
        for handle in hook_handles:
            handle.remove()

    def _maybe_adjust_loss_scale(self, loss: Tensor, output: Tensor) -> Tensor:
        """Adjust the scale of the loss tensor if necessary.

        The ``BCEWithLogitsLoss`` and ``MSELoss`` also average over the output dimension
        in addition to the batch dimension. We adjust the scale of the loss to correct
        for this.

        Args:
            loss: The loss tensor to adjust.
            output: The model's output.

        Returns:
            The scaled loss tensor.
        """
        if (
            isinstance(self._loss_func, (BCEWithLogitsLoss, MSELoss))
            and self._loss_func.reduction == "mean"
        ):
            # ``BCEWithLogitsLoss`` and ``MSELoss`` also average over non-batch
            # dimensions. We have to scale the loss to incorporate this scaling.
            _, C = output.shape
            loss *= sqrt(C)
        return loss

    def _compute_loss_and_backward(self, output: Tensor, y: Tensor):
        r"""Compute the loss and the backward pass(es) required for KFAC.

        Args:
            output: The model's prediction
                :math:`\{f_\mathbf{\theta}(\mathbf{x}_n)\}_{n=1}^N`.
            y: The labels :math:`\{\mathbf{y}_n\}_{n=1}^N`.

        Raises:
            ValueError: If ``fisher_type`` is not ``FisherType.TYPE2``,
                ``FisherType.MC``, ``FisherType.EMPIRICAL``, or
                ``FisherType.FORWARD_ONLY``.
        """
        # if >2d output we convert to an equivalent 2d output
        if isinstance(self._loss_func, CrossEntropyLoss):
            output = rearrange(output, "batch c ... -> (batch ...) c")
            y = rearrange(y, "batch ... -> (batch ...)")
        else:
            output = rearrange(output, "batch ... c -> (batch ...) c")
            y = rearrange(y, "batch ... c -> (batch ...) c")

        if self._fisher_type == FisherType.TYPE2:
            # Compute per-sample Hessian square root, then concatenate over samples.
            # Result has shape `(batch_size, num_classes, num_classes)`
            hessian_sqrts = stack(
                [
                    loss_hessian_matrix_sqrt(out.detach(), target, self._loss_func)
                    for out, target in zip(output.split(1), y.split(1))
                ]
            )

            # Fix scaling caused by the batch dimension
            num_loss_terms = output.shape[0]
            reduction = self._loss_func.reduction
            scale = {"sum": 1.0, "mean": 1.0 / num_loss_terms}[reduction]
            hessian_sqrts.mul_(scale)

            # For each column `c` of the matrix square root we need to backpropagate,
            # but we can do this for all samples in parallel
            num_cols = hessian_sqrts.shape[-1]
            for c in range(num_cols):
                batched_column = hessian_sqrts[:, :, c]
                grad(
                    (output * batched_column).sum(),
                    self._params,
                    retain_graph=c < num_cols - 1,
                )

        elif self._fisher_type == FisherType.MC:
            for mc in range(self._mc_samples):
                y_sampled = self.draw_label(output)
                loss = self._loss_func(output, y_sampled)
                loss = self._maybe_adjust_loss_scale(loss, output)
                grad(loss, self._params, retain_graph=mc != self._mc_samples - 1)

        elif self._fisher_type == FisherType.EMPIRICAL:
            loss = self._loss_func(output, y)
            loss = self._maybe_adjust_loss_scale(loss, output)
            grad(loss, self._params)

        elif self._fisher_type == FisherType.FORWARD_ONLY:
            # Since FOOF sets the gradient covariance Kronecker factors to the identity,
            # we don't need to do a backward pass. See https://arxiv.org/abs/2201.12250.
            # We choose to set the gradient covariance to the identity explicitly for
            # the sake of simplicity, such that the rest of the code here and for
            # `KFACInverseLinearOperator` does not have to be adapted. This could be
            # changed to decrease the memory costs.
            for mod_name, param_pos in self._mapping.items():
                # We iterate over _mapping to get the module names corresponding to the
                # parameters. We only need the output dimension of the module, but
                # don't know whether the parameter is a weight or bias; therefore, we
                # just call `next(iter(param_pos.values()))` to get the first parameter.
                param = self._params[next(iter(param_pos.values()))]
                self._gradient_covariances[mod_name] = eye(
                    param.shape[0], dtype=param.dtype, device=self._device
                )

        else:
            raise ValueError(
                f"Invalid fisher_type: {self._fisher_type}. "
                + f"Supported: {self._SUPPORTED_FISHER_TYPE}."
            )

    def draw_label(self, output: Tensor) -> Tensor:
        r"""Draw a sample from the model's predictive distribution.

        The model's distribution is implied by the (negative log likelihood) loss
        function. For instance, ``MSELoss`` implies a Gaussian distribution with
        constant variance, and ``CrossEntropyLoss`` implies a categorical distribution.

        Args:
            output: The model's prediction
                :math:`\{f_\mathbf{\theta}(\mathbf{x}_n)\}_{n=1}^N`.

        Returns:
            A sample
            :math:`\{\mathbf{y}_n\}_{n=1}^N` drawn from the model's predictive
            distribution :math:`p(\mathbf{y} \mid \mathbf{x}, \mathbf{\theta})`. Has
            the same shape as the labels that would be fed into the loss function
            together with ``output``.

        Raises:
            ValueError: If the output is not 2d.
            NotImplementedError: If the loss function is not supported.
        """
        if output.ndim != 2:
            raise ValueError("Only a 2d output is supported.")

        if isinstance(self._loss_func, MSELoss):
            std = sqrt(0.5)
            perturbation = std * randn(
                output.shape,
                device=output.device,
                dtype=output.dtype,
                generator=self._generator,
            )
            return output.clone().detach() + perturbation

        elif isinstance(self._loss_func, CrossEntropyLoss):
            probs = output.softmax(dim=1)
            labels = probs.multinomial(
                num_samples=1, generator=self._generator
            ).squeeze(-1)
            return labels

        elif isinstance(self._loss_func, BCEWithLogitsLoss):
            probs = output.sigmoid()
            labels = probs.bernoulli(generator=self._generator)
            return labels

        else:
            raise NotImplementedError

    def _register_tensor_hook_on_output_to_accumulate_gradient_covariance(
        self, module: Module, inputs: Tuple[Tensor], output: Tensor, module_name: str
    ):
        """Register tensor hook on layer's output to accumulate the grad. covariance.

        Note:
            The easier way to compute the gradient covariance would be via a full
            backward hook on the module itself which performs the computation.
            However, this approach breaks down if the output of a layer feeds into an
            activation with `inplace=True` (see
            https://github.com/pytorch/pytorch/issues/61519). Hence we use the
            workaround
            https://github.com/pytorch/pytorch/issues/61519#issuecomment-883524237, and
            install a module hook which installs a tensor hook on the module's output
            tensor, which performs the accumulation of the gradient covariance.

        Args:
            module: Layer onto whose output a tensor hook to accumulate the gradient
                covariance will be installed.
            inputs: The layer's input tensors.
            output: The layer's output tensor.
            module_name: The name of the layer in the neural network.
        """
        tensor_hook = partial(
            self._accumulate_gradient_covariance, module=module, module_name=module_name
        )
        output.register_hook(tensor_hook)

    def _accumulate_gradient_covariance(
        self, grad_output: Tensor, module: Module, module_name: str
    ):
        """Accumulate the gradient covariance for a layer's output.

        Updates ``self._gradient_covariances``.

        Args:
            grad_output: The gradient w.r.t. the output.
            module: The layer whose output's gradient covariance will be accumulated.
            module_name: The name of the layer in the neural network.
        """
        g = grad_output.data.detach()
        batch_size = g.shape[0]
        if isinstance(module, Conv2d):
            g = rearrange(g, "batch c o1 o2 -> batch o1 o2 c")

        if self._kfac_approx == KFACType.EXPAND:
            # KFAC-expand approximation
            g = rearrange(g, "batch ... d_out -> (batch ...) d_out")
        else:
            # KFAC-reduce approximation
            g = reduce(g, "batch ... d_out -> batch d_out", "sum")

        # Compute correction for the loss scaling depending on the loss reduction used
        num_loss_terms = batch_size * self._num_per_example_loss_terms
        # self._mc_samples will be 1 if fisher_type != FisherType.MC
        correction = {
            "sum": 1.0 / self._mc_samples,
            "mean": num_loss_terms**2
            / (self._N_data * self._mc_samples * self._num_per_example_loss_terms),
        }[self._loss_func.reduction]

        covariance = einsum(g, g, "b i,b j->i j").mul_(correction)

        if module_name not in self._gradient_covariances:
            self._gradient_covariances[module_name] = covariance
        else:
            self._gradient_covariances[module_name].add_(covariance)

    def _hook_accumulate_input_covariance(
        self, module: Module, inputs: Tuple[Tensor], module_name: str
    ):
        """Pre-forward hook that accumulates the input covariance of a layer.

        Updates ``self._input_covariances``.

        Args:
            module: Module on which the hook is called.
            inputs: Inputs to the module.
            module_name: Name of the module in the neural network.

        Raises:
            ValueError: If the module has multiple inputs.
        """
        if len(inputs) != 1:
            raise ValueError("Modules with multiple inputs are not supported.")
        x = inputs[0].data.detach()

        if isinstance(module, Conv2d):
            patch_extractor_fn = {
                KFACType.EXPAND: extract_patches,
                KFACType.REDUCE: extract_averaged_patches,
            }[self._kfac_approx]
            x = patch_extractor_fn(
                x,
                module.kernel_size,
                module.stride,
                module.padding,
                module.dilation,
                module.groups,
            )

        if self._kfac_approx == KFACType.EXPAND:
            # KFAC-expand approximation
            scale = x.shape[1:-1].numel()  # sequence length
            x = rearrange(x, "batch ... d_in -> (batch ...) d_in")
        else:
            # KFAC-reduce approximation
            scale = 1.0  # since we use a mean reduction
            x = reduce(x, "batch ... d_in -> batch d_in", "mean")

        params = self._mapping[module_name]
        if (
            "weight" in params.keys()
            and "bias" in params.keys()
            and not self._separate_weight_and_bias
        ):
            x = cat([x, x.new_ones(x.shape[0], 1)], dim=1)

        covariance = einsum(x, x, "b i,b j -> i j").div_(self._N_data * scale)

        if module_name not in self._input_covariances:
            self._input_covariances[module_name] = covariance
        else:
            self._input_covariances[module_name].add_(covariance)

    @classmethod
    def compute_parameter_mapping(
        cls, params: List[Union[Tensor, Parameter]], model_func: Module
    ) -> Dict[str, Dict[str, int]]:
        """Construct the mapping between layers, their parameters, and positions.

        Args:
            params: List of parameters.
            model_func: The model function.

        Returns:
            A dictionary of dictionaries. The outer dictionary's keys are the names of
            the layers that contain parameters. The interior dictionary's keys are the
            parameter names, and the values their respective positions.

        Raises:
            NotImplementedError: If parameters are found outside supported layers.
        """
        param_ids = [p.data_ptr() for p in params]
        positions = {}
        processed = set()

        for mod_name, mod in model_func.named_modules():
            if isinstance(mod, cls._SUPPORTED_MODULES) and any(
                p.data_ptr() in param_ids for p in mod.parameters()
            ):
                param_positions = {}
                for p_name, p in mod.named_parameters():
                    p_id = p.data_ptr()
                    if p_id in param_ids:
                        pos = param_ids.index(p_id)
                        param_positions[p_name] = pos
                        processed.add(p_id)
                positions[mod_name] = param_positions

        # check that all parameters are in known modules
        if len(processed) != len(param_ids):
            raise NotImplementedError("Found parameters in un-supported layers.")

        return positions

    @property
    def trace(self) -> Tensor:
        r"""Trace of the KFAC approximation.

        Will call ``_compute_kfac`` if it has not been called before and will cache the
        trace until ``_compute_kfac`` is called again. Uses the property of the
        Kronecker product that
        :math:`\text{tr}(A \otimes B) = \text{tr}(A) \text{tr}(B)`.

        Returns:
            Trace of the KFAC approximation.
        """
        if self._trace is not None:
            return self._trace

        if not self._input_covariances and not self._gradient_covariances:
            self._compute_kfac()

        self._trace = 0.0
        for mod_name, param_pos in self._mapping.items():
            tr_ggT = self._gradient_covariances[mod_name].trace()
            if (
                not self._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                self._trace += self._input_covariances[mod_name].trace() * tr_ggT
            else:
                for p_name in param_pos.keys():
                    self._trace += tr_ggT * (
                        self._input_covariances[mod_name].trace()
                        if p_name == "weight"
                        else 1
                    )
        return self._trace

    @property
    def det(self) -> Tensor:
        r"""Determinant of the KFAC approximation.

        Will call ``_compute_kfac`` if it has not been called before and will cache the
        determinant until ``_compute_kfac`` is called again. Uses the property of the
        Kronecker product that :math:`\det(A \otimes B) = \det(A)^{m} \det(B)^{n}`,
        where
        :math:`A \in \mathbb{R}^{n \times n}` and :math:`B \in \mathbb{R}^{m \times m}`.

        Returns:
            Determinant of the KFAC approximation.
        """
        if self._det is not None:
            return self._det

        if not self._input_covariances and not self._gradient_covariances:
            self._compute_kfac()

        self._det = 1.0
        for mod_name, param_pos in self._mapping.items():
            m = self._gradient_covariances[mod_name].shape[0]
            det_ggT = self._gradient_covariances[mod_name].det()
            if (
                not self._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                n = self._input_covariances[mod_name].shape[0]
                det_aaT = self._input_covariances[mod_name].det()
                self._det *= det_aaT.pow(m) * det_ggT.pow(n)
            else:
                for p_name in param_pos.keys():
                    n = (
                        self._input_covariances[mod_name].shape[0]
                        if p_name == "weight"
                        else 1
                    )
                    self._det *= det_ggT.pow(n) * (
                        self._input_covariances[mod_name].det().pow(m)
                        if p_name == "weight"
                        else 1
                    )
        return self._det

    @property
    def logdet(self) -> Tensor:
        r"""Log determinant of the KFAC approximation.

        More numerically stable than the ``det`` property.
        Will call ``_compute_kfac`` if it has not been called before and will cache the
        log determinant until ``_compute_kfac`` is called again. Uses the property of
        the Kronecker product that
        :math:`\log \det(A \otimes B) = m \log \det(A) + n \log \det(B)`, where
        :math:`A \in \mathbb{R}^{n \times n}` and :math:`B \in \mathbb{R}^{m \times m}`.

        Returns:
            Log determinant of the KFAC approximation.
        """
        if self._logdet is not None:
            return self._logdet

        if not self._input_covariances and not self._gradient_covariances:
            self._compute_kfac()

        self._logdet = 0.0
        for mod_name, param_pos in self._mapping.items():
            m = self._gradient_covariances[mod_name].shape[0]
            logdet_ggT = self._gradient_covariances[mod_name].logdet()
            if (
                not self._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                n = self._input_covariances[mod_name].shape[0]
                logdet_aaT = self._input_covariances[mod_name].logdet()
                self._logdet += m * logdet_aaT + n * logdet_ggT
            else:
                for p_name in param_pos.keys():
                    n = (
                        self._input_covariances[mod_name].shape[0]
                        if p_name == "weight"
                        else 1
                    )
                    self._logdet += n * logdet_ggT + (
                        m * self._input_covariances[mod_name].logdet()
                        if p_name == "weight"
                        else 0
                    )
        return self._logdet

    @property
    def frobenius_norm(self) -> Tensor:
        r"""Frobenius norm of the KFAC approximation.

        Will call ``_compute_kfac`` if it has not been called before and will cache the
        Frobenius norm until ``_compute_kfac`` is called again. Uses the property of the
        Kronecker product that :math:`\|A \otimes B\|_F = \|A\|_F \|B\|_F`.

        Returns:
            Frobenius norm of the KFAC approximation.
        """
        if self._frobenius_norm is not None:
            return self._frobenius_norm

        if not self._input_covariances and not self._gradient_covariances:
            self._compute_kfac()

        self._frobenius_norm = 0.0
        for mod_name, param_pos in self._mapping.items():
            squared_frob_ggT = self._gradient_covariances[mod_name].square().sum()
            if (
                not self._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                squared_frob_aaT = self._input_covariances[mod_name].square().sum()
                self._frobenius_norm += squared_frob_aaT * squared_frob_ggT
            else:
                for p_name in param_pos.keys():
                    self._frobenius_norm += squared_frob_ggT * (
                        self._input_covariances[mod_name].square().sum()
                        if p_name == "weight"
                        else 1
                    )
        self._frobenius_norm.sqrt_()
        return self._frobenius_norm

    def state_dict(self) -> Dict[str, Any]:
        """Return the state of the KFAC linear operator.

        Returns:
            State dictionary.
        """
        loss_type = {
            MSELoss: "MSELoss",
            CrossEntropyLoss: "CrossEntropyLoss",
            BCEWithLogitsLoss: "BCEWithLogitsLoss",
        }[type(self._loss_func)]
        return {
            # Model and loss function
            "model_func_state_dict": self._model_func.state_dict(),
            "loss_type": loss_type,
            "loss_reduction": self._loss_func.reduction,
            # Attributes
            "progressbar": self._progressbar,
            "shape": self.shape,
            "seed": self._seed,
            "fisher_type": self._fisher_type,
            "mc_samples": self._mc_samples,
            "kfac_approx": self._kfac_approx,
            "num_per_example_loss_terms": self._num_per_example_loss_terms,
            "separate_weight_and_bias": self._separate_weight_and_bias,
            "num_data": self._N_data,
            # Kronecker factors (if computed)
            "input_covariances": self._input_covariances,
            "gradient_covariances": self._gradient_covariances,
            # Properties (not necessarily computed)
            "trace": self._trace,
            "det": self._det,
            "logdet": self._logdet,
            "frobenius_norm": self._frobenius_norm,
        }

    def load_state_dict(self, state_dict: Dict[str, Any]):
        """Load the state of the KFAC linear operator.

        Args:
            state_dict: State dictionary.

        Raises:
            ValueError: If the loss function does not match the state dict.
            ValueError: If the loss function reduction does not match the state dict.
        """
        self._model_func.load_state_dict(state_dict["model_func_state_dict"])
        # Verify that the loss function and its reduction match the state dict
        loss_func_type = {
            "MSELoss": MSELoss,
            "CrossEntropyLoss": CrossEntropyLoss,
            "BCEWithLogitsLoss": BCEWithLogitsLoss,
        }[state_dict["loss_type"]]
        if not isinstance(self._loss_func, loss_func_type):
            raise ValueError(
                f"Loss function mismatch: {loss_func_type} != {type(self._loss_func)}."
            )
        if state_dict["loss_reduction"] != self._loss_func.reduction:
            raise ValueError(
                "Loss function reduction mismatch: "
                f"{state_dict['loss_reduction']} != {self._loss_func.reduction}."
            )

        # Set attributes
        self._progressbar = state_dict["progressbar"]
        self.shape = state_dict["shape"]
        self._seed = state_dict["seed"]
        self._fisher_type = state_dict["fisher_type"]
        self._mc_samples = state_dict["mc_samples"]
        self._kfac_approx = state_dict["kfac_approx"]
        self._num_per_example_loss_terms = state_dict["num_per_example_loss_terms"]
        self._separate_weight_and_bias = state_dict["separate_weight_and_bias"]
        self._N_data = state_dict["num_data"]

        # Set Kronecker factors (if computed)
        if self._input_covariances or self._gradient_covariances:
            # If computed, check if the keys match the mapping keys
            input_covariances_keys = set(self._input_covariances.keys())
            gradient_covariances_keys = set(self._gradient_covariances.keys())
            mapping_keys = set(self._mapping.keys())
            if (
                input_covariances_keys != mapping_keys
                or gradient_covariances_keys != mapping_keys
            ):
                raise ValueError(
                    "Input or gradient covariance keys in state dict do not match "
                    "mapping keys of linear operator. "
                    "Difference between input covariance and mapping keys: "
                    f"{input_covariances_keys - mapping_keys}. "
                    "Difference between gradient covariance and mapping keys: "
                    f"{gradient_covariances_keys - mapping_keys}."
                )
        self._input_covariances = state_dict["input_covariances"]
        self._gradient_covariances = state_dict["gradient_covariances"]

        # Set properties (not necessarily computed)
        self._trace = state_dict["trace"]
        self._det = state_dict["det"]
        self._logdet = state_dict["logdet"]
        self._frobenius_norm = state_dict["frobenius_norm"]

    @classmethod
    def from_state_dict(
        cls,
        state_dict: Dict[str, Any],
        model_func: Module,
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        check_deterministic: bool = True,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ) -> KFACLinearOperator:
        """Load a KFAC linear operator from a state dictionary.

        Args:
            state_dict: State dictionary.
            model_func: The model function.
            params: The model's parameters that KFAC is computed for.
            data: A data loader containing the data of the Fisher/GGN.
            check_deterministic: Whether to check that the linear operator is
                deterministic. Defaults to ``True``.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.

        Returns:
            Linear operator of KFAC approximation.

        Raises:
            RuntimeError: If the check for deterministic behavior fails.
        """
        loss_func = {
            "MSELoss": MSELoss,
            "CrossEntropyLoss": CrossEntropyLoss,
            "BCEWithLogitsLoss": BCEWithLogitsLoss,
        }[state_dict["loss_type"]](reduction=state_dict["loss_reduction"])
        kfac = cls(
            model_func,
            loss_func,
            params,
            data,
            batch_size_fn=batch_size_fn,
            check_deterministic=False,
            progressbar=state_dict["progressbar"],
            shape=state_dict["shape"],
            seed=state_dict["seed"],
            fisher_type=state_dict["fisher_type"],
            mc_samples=state_dict["mc_samples"],
            kfac_approx=state_dict["kfac_approx"],
            num_per_example_loss_terms=state_dict["num_per_example_loss_terms"],
            separate_weight_and_bias=state_dict["separate_weight_and_bias"],
            num_data=state_dict["num_data"],
        )
        kfac.load_state_dict(state_dict)

        # Potentially call `check_deterministic` after the state dict is loaded
        if check_deterministic:
            old_device = kfac._device
            kfac.to_device(device("cpu"))
            try:
                kfac._check_deterministic()
            except RuntimeError as e:
                raise e
            finally:
                kfac.to_device(old_device)

        return kfac



"""Contains LinearOperator implementation of gradient moment matrices."""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import Callable, Iterable, List, Optional, Tuple, Union

from backpack.hessianfree.ggnvp import ggn_vector_product_from_plist
from einops import einsum, rearrange
from torch import Tensor, zeros_like
from torch.autograd import grad
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss, Parameter

from curvlinops._base import _LinearOperator


class EFLinearOperator(_LinearOperator):
    r"""Uncentered gradient covariance as SciPy linear operator.

    The uncentered gradient covariance is often called 'empirical Fisher' (EF).

    Consider the empirical risk

    .. math::
        \mathcal{L}(\mathbf{\theta})
        =
        c \sum_{n=1}^{N}
        \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)

    with :math:`c = \frac{1}{N}` for ``reduction='mean'`` and :math:`c=1` for
    ``reduction='sum'``. The uncentered gradient covariance matrix is

    .. math::
        c \sum_{n=1}^{N}
        \left(
            \nabla_{\mathbf{\theta}}
            \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)
        \right)
        \left(
            \nabla_{\mathbf{\theta}}
            \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)
        \right)^\top\,.

    .. note::
        Multiplication with the empirical Fisher is currently implemented with an
        inefficient for-loop.
    """

    supported_losses = (MSELoss, CrossEntropyLoss, BCEWithLogitsLoss)

    def __init__(
        self,
        model_func: Callable[[Tensor], Tensor],
        loss_func: Union[Callable[[Tensor, Tensor], Tensor], None],
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        progressbar: bool = False,
        check_deterministic: bool = True,
        num_data: Optional[int] = None,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ):
        """Linear operator for the uncentered gradient covariance/empirical Fisher (EF).

        Note:
            f(X; θ) denotes a neural network, parameterized by θ, that maps a mini-batch
            input X to predictions p. ℓ(p, y) maps the prediction to a loss, using the
            mini-batch labels y.

        Args:
            model_func: A function that maps the mini-batch input X to predictions.
                Could be a PyTorch module representing a neural network.
            loss_func: Loss function criterion. Maps predictions and mini-batch labels
                to a scalar value.
            params: List of differentiable parameters used by the prediction function.
            data: Source from which mini-batches can be drawn, for instance a list of
                mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``. Note that ``X``
                could be a ``dict`` or ``UserDict``; this is useful for custom models.
                In this case, you must (i) specify the ``batch_size_fn`` argument, and
                (ii) take care of preprocessing like ``X.to(device)`` inside of your
                ``model.forward()`` function. Due to the sequential internal Monte-Carlo
                sampling, batches must be presented in the same deterministic
                order (no shuffling!).
            progressbar: Show a progressbar during matrix-multiplication.
                Default: ``False``.
            check_deterministic: Probe that model and data are deterministic, i.e.
                that the data does not use `drop_last` or data augmentation. Also, the
                model's forward pass could depend on the order in which mini-batches
                are presented (BatchNorm, Dropout). Default: ``True``. This is a
                safeguard, only turn it off if you know what you are doing.
            num_data: Number of data points. If ``None``, it is inferred from the data
                at the cost of one traversal through the data loader.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.

        Raises:
             NotImplementedError: If the loss function differs from ``MSELoss``,
                 BCEWithLogitsLoss, or ``CrossEntropyLoss``.
        """
        if not isinstance(loss_func, self.supported_losses):
            raise NotImplementedError(
                f"Loss must be one of {self.supported_losses}. Got: {loss_func}."
            )
        super().__init__(
            model_func,
            loss_func,
            params,
            data,
            progressbar=progressbar,
            check_deterministic=check_deterministic,
            num_data=num_data,
            batch_size_fn=batch_size_fn,
        )

    def _matmat_batch(
        self, X: Union[Tensor, MutableMapping], y: Tensor, M_list: List[Tensor]
    ) -> Tuple[Tensor, ...]:
        """Apply the mini-batch empirical Fisher to a matrix.

        Args:
            X: Input to the DNN.
            y: Ground truth.
            M_list: Matrix to be multiplied with in list format.
                Tensors have same shape as trainable model parameters, and an
                additional leading axis for the matrix columns.

        Returns:
            Result of EF multiplication in list format. Has the same shape as
            ``M_list``, i.e. each tensor in the list has the shape of a parameter and a
            leading dimension of matrix columns.
        """
        output = self._model_func(X)
        # If >2d output we convert to an equivalent 2d output
        if isinstance(self._loss_func, CrossEntropyLoss):
            output = rearrange(output, "batch c ... -> (batch ...) c")
            y = rearrange(y, "batch ... -> (batch ...)")
        else:
            output = rearrange(output, "batch ... c -> (batch ...) c")
            y = rearrange(y, "batch ... c -> (batch ...) c")

        # Adjust the scale depending on the loss reduction used
        num_loss_terms, C = output.shape
        reduction_factor = {
            "mean": (
                num_loss_terms
                if isinstance(self._loss_func, CrossEntropyLoss)
                else num_loss_terms * C
            ),
            "sum": 1.0,
        }[self._loss_func.reduction]

        # compute ∂ℓₙ/∂fₙ without reduction factor of L
        (grad_output,) = grad(self._loss_func(output, y), output)
        grad_output = grad_output.detach() * reduction_factor

        # Compute the pseudo-loss L' := 0.5 / c ∑ₙ fₙᵀ (gₙ gₙᵀ) fₙ where gₙ = ∂ℓₙ/∂fₙ
        # (detached). The GGN of L' linearized at fₙ is the empirical Fisher.
        # We can thus multiply with the EF by computing the GGN-vector products of L'.
        loss = (
            0.5
            / reduction_factor
            * (einsum(output, grad_output, "n ..., n ... -> n") ** 2).sum()
        )

        # Multiply the EF onto each vector in the input matrix
        result_list = [zeros_like(M) for M in M_list]
        num_vectors = M_list[0].shape[0]
        for v in range(num_vectors):
            for idx, ggnvp in enumerate(
                ggn_vector_product_from_plist(
                    loss, output, self._params, [M[v] for M in M_list]
                )
            ):
                result_list[idx][v].add_(ggnvp.detach())

        return tuple(result_list)

    def _adjoint(self) -> EFLinearOperator:
        """Return the linear operator representing the adjoint.

        The empirical Fisher is real symmetric, and hence self-adjoint.

        Returns:
            Self.
        """
        return self



"""Hutchinson-style matrix diagonal estimation."""

from numpy import ndarray
from scipy.sparse.linalg import LinearOperator

from curvlinops.sampling import random_vector


class HutchinsonDiagonalEstimator:
    r"""Class to perform diagonal estimation with Hutchinson's method.

    For details, see

    - Martens, J., Sutskever, I., & Swersky, K. (2012). Estimating the hessian by
      back-propagating curvature. International Conference on Machine Learning (ICML).

    Let :math:`\mathbf{A}` be a square linear operator. We can approximate its diagonal
    :math:`\mathrm{diag}(\mathbf{A})` by drawing a random vector :math:`\mathbf{v}`
    which satisfies :math:`\mathbb{E}[\mathbf{v} \mathbf{v}^\top] = \mathbf{I}` and
    sample from the estimator

    .. math::
        \mathbf{a}
        := \mathbf{v} \odot \mathbf{A} \mathbf{v}
        \approx \mathrm{diag}(\mathbf{A})\,.

    This estimator is unbiased,

    .. math::
        \mathbb{E}[a_i]
        = \sum_j \mathbb{E}[v_i A_{i,j} v_j]
        = \sum_j A_{i,j} \mathbb{E}[v_i  v_j]
        = \sum_j A_{i,j} \delta_{i, j}
        = A_{i,i}\,.

    Example:
        >>> from numpy import diag, mean, round
        >>> from numpy.random import rand, seed
        >>> from numpy.linalg import norm
        >>> seed(0) # make deterministic
        >>> A = rand(10, 10)
        >>> diag_A = diag(A) # exact diagonal as reference
        >>> estimator = HutchinsonDiagonalEstimator(A)
        >>> # one- and multi-sample approximations
        >>> diag_A_low_precision = estimator.sample()
        >>> samples = [estimator.sample() for _ in range(1_000)]
        >>> diag_A_high_precision = mean(samples, axis=0)
        >>> # compute residual norms
        >>> error_low_precision = norm(diag_A - diag_A_low_precision)
        >>> error_high_precision = norm(diag_A - diag_A_high_precision)
        >>> assert error_low_precision > error_high_precision
        >>> round(error_low_precision, 4), round(error_high_precision, 4)
        (5.7268, 0.1525)
    """

    def __init__(self, A: LinearOperator):
        """Store the linear operator whose diagonal will be estimated.

        Args:
            A: Linear square-shaped operator whose diagonal will be estimated.

        Raises:
            ValueError: If the operator is not square.
        """
        if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
            raise ValueError(f"A must be square. Got shape {A.shape}.")
        self._A = A

    def sample(self, distribution: str = "rademacher") -> ndarray:
        """Draw a sample from the diagonal estimator.

        Multiple samples can be combined into a more accurate diagonal estimation via
        averaging.

        Args:
            distribution: Distribution of the vector along which the linear operator
                will be evaluated. Either ``'rademacher'`` or ``'normal'``.
                Default is ``'rademacher'``.

        Returns:
            A Sample from the diagonal estimator.
        """
        dim = self._A.shape[1]
        v = random_vector(dim, distribution)
        Av = self._A @ v
        return v * Av



"""Spectral analysis methods for SciPy linear operators.

From Papyan, 2020:

- Traces of class/cross-class structure pervade deep learning spectra. Journal
  of Machine Learning Research (JMLR), https://jmlr.org/papers/v21/20-933.html
"""

from typing import List, Tuple, Union

from numpy import exp, inner, linspace, log, ndarray, pi, sqrt, zeros, zeros_like
from numpy.linalg import norm
from numpy.random import randn
from scipy.linalg import eigh, eigh_tridiagonal
from scipy.sparse import diags
from scipy.sparse.linalg import LinearOperator, eigsh


def lanczos_approximate_spectrum(
    A: LinearOperator,
    ncv: int,
    num_points: int = 1024,
    num_repeats: int = 1,
    kappa: float = 3.0,
    boundaries: Union[
        Tuple[float, float], Tuple[float, None], Tuple[None, float], None
    ] = None,
    margin: float = 0.05,
    boundaries_tol: float = 1e-2,
) -> Tuple[ndarray, ndarray]:
    """Approximate the spectral density p(λ) = 1/d ∑ᵢ δ(λ - λᵢ) of A ∈ Rᵈˣᵈ.

    Implements algorithm 2 (:code:`LanczosApproxSpec`) of Papyan, 2020
    (https://jmlr.org/papers/v21/20-933.html).

    Internally rescales the operator spectrum to the interval [-1; 1] such that
    the width ``kappa`` of the Gaussian bumps used to approximate the delta peaks
    need not be tweaked.

    Args:
        A: Symmetric linear operator.
        ncv: Number of Lanczos vectors (number of nodes/weights for the quadrature).
        num_points: Resolution.
        num_repeats: Number of Lanczos quadratures to average the density over.
            Default: ``1``. Taken from papyan2020traces, Section D.2.
        kappa: Width of the Gaussian used to approximate delta peaks in [-1; 1]. Must
            be greater than 1. Default: ``3``. Taken from papyan2020traces, Section D.2.
        boundaries: Estimates of the minimum and maximum eigenvalues of ``A``. If left
            unspecified, they will be estimated internally.
        margin: Relative margin added around the spectral boundary. Default: ``0.05``.
            Taken from papyan2020traces, Section D.2.
        boundaries_tol: (Only relevant if ``boundaries`` are not specified). Relative
            accuracy used to estimate the spectral boundary. ``0`` implies machine
            precision. Default: ``1e-2``, from
            https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.

    Returns:
        Grid points λ and approximated spectral density p(λ) of A.
    """
    boundaries = approximate_boundaries(A, tol=boundaries_tol, boundaries=boundaries)

    average_density = zeros(num_points)

    for n in range(num_repeats):
        lanczos_iter = fast_lanczos(A, ncv)
        grid, density = lanczos_approximate_spectrum_from_iter(
            lanczos_iter, boundaries, num_points, kappa, margin
        )

        average_density = (1 - 1 / (n + 1)) * average_density + density / (n + 1)

    return grid, average_density


def lanczos_approximate_spectrum_from_iter(
    lanczos_iter: Tuple[ndarray, ndarray],
    boundaries: Tuple[float, float],
    num_points: int,
    kappa: float,
    margin: float,
) -> Tuple[ndarray, ndarray]:
    eval_min, eval_max = boundaries
    _width = eval_max - eval_min
    _padding = margin * _width
    eval_min, eval_max = eval_min - _padding, eval_max + _padding

    # use normalized operator ``(A - c I) / d`` whose spectrum lies in [-1; 1]
    c = (eval_max + eval_min) / 2
    d = (eval_max - eval_min) / 2

    # estimate on grid [-1; 1]
    grid_norm = linspace(-1, 1, num_points, endpoint=True)
    density = zeros_like(grid_norm)

    evals, evecs = lanczos_iter
    ncv = evals.shape[0]
    nodes = (evals - c) / d
    # Repeat as ``(ncv, num_points)`` arrays to avoid broadcasting
    grid = grid_norm.reshape((1, num_points)).repeat(ncv, axis=0)
    nodes = nodes.reshape((ncv, 1)).repeat(num_points, axis=1)
    weights = (evecs[0, :] ** 2 / d).reshape((ncv, 1)).repeat(num_points, axis=1)

    # width of Gaussian bump in [-1; 1]
    sigma = 2 / (ncv - 1) / sqrt(8 * log(kappa))
    density = (weights * _gaussian(grid, nodes, sigma)).sum(0)

    return linspace(eval_min, eval_max, num_points, endpoint=True), density


class _LanczosSpectrumCached:
    """Base class for approximating spectra with Lanczos iterations.

    Caches the Lanczos iterations to efficiently produce approximations with different
    hyperparameters.
    """

    def __init__(self, A: LinearOperator, ncv: int):
        """Initialize.

        Args:
            A: Symmetric linear operator.
            ncv: Number of Lanczos vectors (number of nodes/weights for the quadrature).
        """
        self._A = A
        self._ncv = ncv
        self._lanczos_iters: List[Tuple[ndarray, ndarray]] = []

    def _get_lanczos_iters(self, num_iters: int) -> List[Tuple[ndarray, ndarray]]:
        while len(self._lanczos_iters) < num_iters:
            self._lanczos_iters.append(fast_lanczos(self._A, self._ncv))

        return self._lanczos_iters[:num_iters]


class LanczosApproximateSpectrumCached(_LanczosSpectrumCached):
    """Class to approximate the spectral density of p(λ) = 1/d ∑ᵢ δ(λ - λᵢ) of A ∈ Rᵈˣᵈ.

    Caches Lanczos iterations to efficiently produce spectral density approximations with
    different hyperparameters.
    """

    def __init__(
        self,
        A: LinearOperator,
        ncv: int,
        boundaries: Union[
            Tuple[float, float], Tuple[float, None], Tuple[None, float], None
        ] = None,
        boundaries_tol: float = 1e-2,
    ):
        """Initialize.

        Args:
            A: Symmetric linear operator.
            ncv: Number of Lanczos vectors (number of nodes/weights for the quadrature).
            boundaries: Estimates of the minimum and maximum eigenvalues of ``A``. If
                left unspecified, they will be estimated internally.
            boundaries_tol: (Only relevant if ``boundaries`` are not specified).
                Relative accuracy used to estimate the spectral boundary. ``0`` implies
                machine precision. Default: ``1e-2``, from
                https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.
        """
        super().__init__(A, ncv)
        self._boundaries = approximate_boundaries(
            A, tol=boundaries_tol, boundaries=boundaries
        )

    def approximate_spectrum(
        self,
        num_repeats: int = 1,
        num_points: int = 1024,
        kappa: float = 3.0,
        margin: float = 0.05,
    ) -> Tuple[ndarray, ndarray]:
        """Approximate the spectal density of A.

        Args:
            num_repeats: Number of Lanczos quadratures to average the density over.
                Default: ``1``. Taken from papyan2020traces, Section D.2.
            num_points: Resolution. Default: ``1024``.
            kappa: Width of the Gaussian used to approximate delta peaks in [-1; 1]. Must
                be greater than 1. Default: ``3``. From papyan2020traces, Section D.2.
            margin: Relative margin added around the spectral boundary.
                Default: ``0.05``. Taken from papyan2020traces, Section D.2.

        Returns:
            Grid points λ and approximated spectral density p(λ) of A.
        """
        spectra = [
            lanczos_approximate_spectrum_from_iter(
                lanczos_iter, self._boundaries, num_points, kappa, margin
            )
            for lanczos_iter in self._get_lanczos_iters(num_repeats)
        ]
        grid = spectra[0][0]
        spectrum = sum(spectrum[1] for spectrum in spectra) / num_repeats

        return grid, spectrum


def lanczos_approximate_log_spectrum(
    A: LinearOperator,
    ncv: int,
    num_points: int = 1024,
    num_repeats: int = 1,
    kappa: float = 1.04,
    boundaries: Union[
        Tuple[float, float], Tuple[float, None], Tuple[None, float], None
    ] = None,
    margin: float = 0.05,
    boundaries_tol: float = 1e-2,
    epsilon: float = 1e-5,
) -> Tuple[ndarray, ndarray]:
    """Approximate the spectral density p(λ) = 1/d ∑ᵢ δ(λ - λᵢ) of log(|A| + εI) ∈ Rᵈˣᵈ.

    Follows the idea of Section C.7 in Papyan, 2020
    (https://jmlr.org/papers/v21/20-933.html).

    Here, log denotes the natural logarithm (i.e. base e).

    Internally rescales the operator spectrum to the interval [-1; 1] such that
    the width ``kappa`` of the Gaussian bumps used to approximate the delta peaks
    need not be tweaked.

    Args:
        A: Symmetric linear operator.
        ncv: Number of Lanczos vectors (number of nodes/weights for the quadrature).
        num_points: Resolution.
        num_repeats: Number of Lanczos quadratures to average the density over.
            Default: ``1``. Taken from papyan2020traces, Section D.2.
        kappa: Width of the Gaussian used to approximate delta peaks in [-1; 1]. Must
            be greater than 1. Default: ``1.04``. Obtained by tweaking while reproducing
            Fig. 15b from papyan2020traces (not specified by the paper).
        boundaries: Estimates of the minimum and maximum eigenvalues of ``|A|``. If left
            unspecified, they will be estimated internally.
        margin: Relative margin added around the spectral boundary. Default: ``0.05``.
            Taken from papyan2020traces, Section D.2.
        boundaries_tol: (Only relevant if ``boundaries`` are not specified). Relative
            accuracy used to estimate the spectral boundary. ``0`` implies machine
            precision. Default: ``1e-2``, from
            https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.
        epsilon: Shift to increase numerical stability. Default: ``1e-5``. Taken from
            papyan2020traces, Section D.2.

    Returns:
        Grid points λ and approximated spectral density p(λ) of log(|A| + εI).
    """
    boundaries = approximate_boundaries_abs(
        A, tol=boundaries_tol, boundaries=boundaries
    )

    average_density = zeros(num_points)

    for n in range(num_repeats):
        lanczos_iter = fast_lanczos(A, ncv)
        grid, density = lanczos_approximate_log_spectrum_from_iter(
            lanczos_iter, boundaries, num_points, kappa, margin, epsilon
        )

        average_density = (1 - 1 / (n + 1)) * average_density + density / (n + 1)

    return grid, average_density


def lanczos_approximate_log_spectrum_from_iter(
    lanczos_iter: Tuple[ndarray, ndarray],
    boundaries: Tuple[float, float],
    num_points: int,
    kappa: float,
    margin: float,
    epsilon: float,
) -> Tuple[ndarray, ndarray]:
    log_eval_min, log_eval_max = (log(boundary + epsilon) for boundary in boundaries)
    _width = log_eval_max - log_eval_min
    _padding = margin * _width
    log_eval_min, log_eval_max = log_eval_min - _padding, log_eval_max + _padding

    # use normalized operator ``(log(|A| + εI) - c I) / d`` with spectrum in [-1; 1]
    c = (log_eval_max + log_eval_min) / 2
    d = (log_eval_max - log_eval_min) / 2

    # estimate on grid [-1; 1]
    grid_norm = linspace(-1, 1, num_points, endpoint=True)
    grid_out = exp(grid_norm * d + c)

    evals, evecs = lanczos_iter

    abs_evals = abs(evals) + epsilon
    log_evals = log(abs_evals)
    nodes = (log_evals - c) / d

    # Repeat as ``(ncv, num_points)`` arrays to avoid broadcasting
    ncv = evals.shape[0]
    grid = grid_norm.reshape((1, num_points)).repeat(ncv, axis=0)
    nodes = nodes.reshape((ncv, 1)).repeat(num_points, axis=1)
    weights = (evecs[0, :] ** 2).reshape((ncv, 1)).repeat(num_points, axis=1)

    # width of Gaussian bump in [-1; 1]
    sigma = 2 / (ncv - 1) / sqrt(8 * log(kappa))
    density = (weights * _gaussian(grid, nodes, sigma)).sum(0) / (d * grid_out)

    return grid_out, density


class LanczosApproximateLogSpectrumCached(_LanczosSpectrumCached):
    """Class to approximate p(λ) = 1/d ∑ᵢ δ(λ - λᵢ) of log(|A| + εI) ∈ Rᵈˣᵈ.

    Caches Lanczos iterations to efficiently produce spectral density approximations with
    different hyperparameters.
    """

    def __init__(
        self,
        A: LinearOperator,
        ncv: int,
        boundaries: Union[
            Tuple[float, float], Tuple[float, None], Tuple[None, float], None
        ] = None,
        boundaries_tol: float = 1e-2,
    ):
        """Initialize.

        Args:
            A: Symmetric linear operator.
            ncv: Number of Lanczos vectors (number of nodes/weights for the quadrature).
            boundaries: Estimates of the minimum and maximum eigenvalues of ``|A|``. If
                left unspecified, they will be estimated internally.
            boundaries_tol: (Only relevant if ``boundaries`` are not specified).
                Relative accuracy used to estimate the spectral boundary. ``0`` implies
                machine precision. Default: ``1e-2``, from
                https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.
        """
        super().__init__(A, ncv)
        self._boundaries = approximate_boundaries_abs(
            A, tol=boundaries_tol, boundaries=boundaries
        )

    def approximate_log_spectrum(
        self,
        num_repeats: int = 1,
        num_points: int = 1024,
        kappa: float = 3.0,
        margin: float = 0.05,
        epsilon: float = 1e-5,
    ) -> Tuple[ndarray, ndarray]:
        """Approximate the spectal density of A.

        Args:
            num_repeats: Number of Lanczos quadratures to average the density over.
                Default: ``1``. Taken from papyan2020traces, Section D.2.
            num_points: Resolution. Default: ``1024``.
            kappa: Width of the Gaussian used to approximate delta peaks in [-1; 1]. Must
                be greater than 1. Default: ``3``. From papyan2020traces, Section D.2.
            margin: Relative margin added around the spectral boundary.
                Default: ``0.05``. Taken from papyan2020traces, Section D.2.
            epsilon: Shift to increase numerical stability. Default: ``1e-5``. Taken from
                papyan2020traces, Section D.2.

        Returns:
            Grid points λ and approximated spectral density p(λ) of log(|A| + εI).
        """
        spectra = [
            lanczos_approximate_log_spectrum_from_iter(
                lanczos_iter, self._boundaries, num_points, kappa, margin, epsilon
            )
            for lanczos_iter in self._get_lanczos_iters(num_repeats)
        ]
        grid = spectra[0][0]
        spectrum = sum(spectrum[1] for spectrum in spectra) / num_repeats

        return grid, spectrum


def fast_lanczos(
    A: LinearOperator, ncv: int, use_eigh_tridiagonal: bool = False
) -> Tuple[ndarray, ndarray]:
    """Lanczos iterations for large-scale problems (no reorthogonalization step).

    Implements algorithm 2 of Papyan, 2020 (https://jmlr.org/papers/v21/20-933.html).

    Args:
        A: Symmetric linear operator.
        ncv: Number of Lanczos vectors.
        use_eigh_tridiagonal: Whether to use eigh_tridiagonal to eigen-decompose the
            tri-diagonal matrix. Default: ``False``. Setting this value to ``True``
            results in faster eigen-decomposition, but is less stable.

    Returns:
        Eigenvalues and eigenvectors of the tri-diagonal matrix built up during
        Lanczos iterations. ``evecs[:, i]`` is normalized eigenvector of ``evals[i]``.
    """
    alphas, betas = zeros(ncv), zeros(ncv - 1)

    dim = A.shape[1]
    v, v_prev = None, None

    for m in range(ncv):
        if m == 0:
            v = randn(dim)
            v /= norm(v)
            v_next = A @ v

        else:
            v_next = A @ v - betas[m - 1] * v_prev

        alphas[m] = inner(v_next, v)
        v_next -= alphas[m] * v

        last = m == ncv - 1
        if not last:
            betas[m] = norm(v_next)
            v_next /= betas[m]
            v_prev = v
            v = v_next

    if use_eigh_tridiagonal:
        evals, evecs = eigh_tridiagonal(alphas, betas)
    else:
        T = diags([betas, alphas, betas], offsets=[-1, 0, 1]).todense()
        evals, evecs = eigh(T)

    return evals, evecs


def approximate_boundaries(
    A: LinearOperator,
    tol: float = 1e-2,
    boundaries: Union[
        Tuple[float, float], Tuple[float, None], Tuple[None, float], None
    ] = None,
) -> Tuple[float, float]:
    """Approximate λₘᵢₙ(A) and λₘₐₓ(A) using SciPy's ``eigsh``.

    Args:
        A: Symmetric linear operator.
        tol: Relative accuracy used by ``eigsh``. ``0`` implies machine precision.
            Default: ``1e-2``, from
            https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.
        boundaries: A tuple of floats that specifies known parts of the boundaries
            which consequently won't be recomputed. Default: ``None``.

    Returns:
        Estimates of λₘᵢₙ and λₘₐₓ.
    """
    eigsh_kwargs = {"tol": tol, "return_eigenvectors": False}

    if boundaries is None:
        eval_min, eval_max = eigsh(A, k=2, which="BE", **eigsh_kwargs)
    else:
        eval_min, eval_max = boundaries

        if eval_min is None:
            (eval_min,) = eigsh(A, k=1, which="SA", **eigsh_kwargs)
        if eval_max is None:
            (eval_max,) = eigsh(A, k=1, which="LA", **eigsh_kwargs)

    return eval_min, eval_max


def approximate_boundaries_abs(
    A: LinearOperator,
    tol: float = 1e-2,
    boundaries: Union[
        Tuple[float, float], Tuple[float, None], Tuple[None, float], None
    ] = None,
) -> Tuple[float, float]:
    """Approximate λₘᵢₙ(|A|) and λₘₐₓ(|A|) using SciPy's ``eigsh``.

    Args:
        A: Symmetric linear operator.
        tol: Relative accuracy used by ``eigsh``. ``0`` implies machine precision.
            Default: ``1e-2``, from
            https://docs.scipy.org/doc/scipy/reference/tutorial/arpack.html#examples.
        boundaries: A tuple of floats that specifies known parts of the boundaries
            which consequently won't be recomputed. Default: ``None``.

    Returns:
        Estimates of λₘᵢₙ and λₘₐₓ of |A|.
    """
    eval_min, eval_max = (None, None) if boundaries is None else boundaries

    eigsh_kwargs = {"tol": tol, "return_eigenvectors": False}
    if eval_max is None:
        (eval_max,) = eigsh(A, k=1, which="LM", **eigsh_kwargs)
    if eval_min is None:
        (eval_min,) = eigsh(A, k=1, which="SM", **eigsh_kwargs)

    return abs(eval_min), abs(eval_max)


def _gaussian(x: ndarray, mu: ndarray, sigma: float) -> ndarray:
    """Normal distribution pdf.

    Args:
        x: Position to evaluate.
        mu: Mean values.
        sigma: Standard deviation.

    Returns:
        Values of normal distribution.
    """
    return exp(-0.5 * ((x - mu) / sigma) ** 2) / (sigma * sqrt(2 * pi))



"""Contains tests for ``curvlinops/jacobian``."""

from collections.abc import MutableMapping

from numpy import random
from pytest import raises

from curvlinops import JacobianLinearOperator, TransposedJacobianLinearOperator
from curvlinops.examples.functorch import functorch_jacobian
from curvlinops.examples.utils import report_nonclose


def test_JacobianLinearOperator_matvec(case, adjoint: bool):
    model_func, _, params, data, batch_size_fn = case

    # Test when X is dict-like but batch_size_fn = None (default)
    if isinstance(data[0][0], MutableMapping):
        with raises(AttributeError):
            op = JacobianLinearOperator(model_func, params, data)

    op = JacobianLinearOperator(model_func, params, data, batch_size_fn=batch_size_fn)
    op_functorch = (
        functorch_jacobian(model_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    x = random.rand(op.shape[1])
    report_nonclose(op @ x, op_functorch @ x, rtol=1e-4)


def test_JacobianLinearOperator_matmat(case, adjoint: bool, num_vecs: int = 3):
    model_func, _, params, data, batch_size_fn = case

    op = JacobianLinearOperator(model_func, params, data, batch_size_fn=batch_size_fn)
    op_functorch = (
        functorch_jacobian(model_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    X = random.rand(op.shape[1], num_vecs)
    report_nonclose(op @ X, op_functorch @ X, rtol=1e-4)


def test_TransposedJacobianLinearOperator_matvec(case, adjoint: bool):
    model_func, _, params, data, batch_size_fn = case

    # Test when X is dict-like but batch_size_fn = None (default)
    if isinstance(data[0][0], MutableMapping):
        with raises(AttributeError):
            op = TransposedJacobianLinearOperator(model_func, params, data)

    op = TransposedJacobianLinearOperator(
        model_func, params, data, batch_size_fn=batch_size_fn
    )
    op_functorch = (
        functorch_jacobian(model_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
        .T
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    x = random.rand(op.shape[1])
    report_nonclose(op @ x, op_functorch @ x, rtol=1e-4)


def test_TransposedJacobianLinearOperator_matmat(
    case, adjoint: bool, num_vecs: int = 3
):
    model_func, _, params, data, batch_size_fn = case

    op = TransposedJacobianLinearOperator(
        model_func, params, data, batch_size_fn=batch_size_fn
    )
    op_functorch = (
        functorch_jacobian(model_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
        .T
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    X = random.rand(op.shape[1], num_vecs)
    report_nonclose(op @ X, op_functorch @ X, rtol=1e-4)



"""Contains tests of ``curvlinops/papyan2020traces/spectrum``."""

from pathlib import Path
from subprocess import CalledProcessError, check_output

from numpy import allclose, array, diag

from curvlinops.papyan2020traces.spectrum import (
    approximate_boundaries,
    approximate_boundaries_abs,
)

PROJECT_ROOT = Path(__file__).parent.parent.parent
BASIC_USAGE = PROJECT_ROOT / "docs" / "examples" / "basic_usage"


def test_example_verification_spectral_density():
    """Integration test to check that the verification example is working.

    It is hard to test the spectral density estimation techniques. This test
    uses the verification example from the documentation.
    """
    EXAMPLE_VERIFICATION_SPECTRAL_DENSITY = (
        BASIC_USAGE / "example_verification_spectral_density.py"
    )

    try:
        check_output(f"python {EXAMPLE_VERIFICATION_SPECTRAL_DENSITY}", shell=True)
    except CalledProcessError as e:
        print(e.output)
        raise e


def test_approximate_boundaries():
    """Test spectrum boundary approximation with partially supplied boundaries."""
    A_diag = array([1.0, 2.0, 3.0, 4.0, 5.0])
    A = diag(A_diag)
    lambda_min, lambda_max = A_diag.min(), A_diag.max()

    cases = [
        [(0.0, 10.0), (0.0, 10.0)],
        [(1.5, None), (1.5, lambda_max)],
        [(None, 2.5), (lambda_min, 2.5)],
        [(None, None), (lambda_min, lambda_max)],
        [None, (lambda_min, lambda_max)],
    ]

    for inputs, results in cases:
        output = approximate_boundaries(A, boundaries=inputs)
        assert len(output) == 2
        assert isinstance(output[0], float)
        assert isinstance(output[1], float)
        assert allclose(output, results)


def test_approximate_boundaries_abs():
    """Test abs spectrum boundary approximation with partially supplied boundaries."""
    A_diag = array([-2.0, -1.0, 3.0, 4.0, 5.0])
    A = diag(A_diag)
    lambda_abs_min, lambda_abs_max = abs(A_diag).min(), abs(A_diag).max()

    cases = [
        [(0.0, 10.0), (0.0, 10.0)],
        [(1.5, None), (1.5, lambda_abs_max)],
        [(None, 2.5), (lambda_abs_min, 2.5)],
        [(None, None), (lambda_abs_min, lambda_abs_max)],
        [None, (lambda_abs_min, lambda_abs_max)],
    ]

    for inputs, results in cases:
        output = approximate_boundaries_abs(A, boundaries=inputs)
        assert len(output) == 2
        assert isinstance(output[0], float)
        assert isinstance(output[1], float)
        assert allclose(output, results)



"""General utility functions."""

from typing import List

from numpy import cumsum


def split_list(x: List, sizes: List[int]) -> List[List]:
    """Split a list into multiple lists of specified size.

    Args:
        x: List to be split.
        sizes: Sizes of the resulting lists.

    Returns:
        List of lists. Each sub-list has the size specified in ``sizes``.

    Raises:
        ValueError: If the sum of ``sizes`` does not match the input list's length.
    """
    if len(x) != sum(sizes):
        raise ValueError(
            f"List to be split has length {len(x)}, but requested sub-list with a total"
            + f" of {sum(sizes)} entries."
        )
    boundaries = cumsum([0] + sizes)
    return [x[boundaries[i] : boundaries[i + 1]] for i in range(len(sizes))]



r"""Usage with Huggingface LLMs
===============================

This example demonstrates how to work with Huggingface (HF) language models.

As always, let's first import the required functionality.
Remember to run :code:`pip install -U transformers datasets`
"""

from collections import UserDict
from collections.abc import MutableMapping

import numpy as np
import torch
import torch.utils.data as data_utils
from datasets import Dataset
from torch import nn
from transformers import (
    DataCollatorWithPadding,
    GPT2Config,
    GPT2ForSequenceClassification,
    GPT2Tokenizer,
    PreTrainedTokenizer,
)

from curvlinops import GGNLinearOperator

# make deterministic
torch.manual_seed(0)
np.random.seed(0)

# %%
#
# Data
# ----
#
# We will use synthetic data for simplicity. But obviously this can
# be replaced with any HF dataloader.

tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
tokenizer.pad_token_id = tokenizer.eos_token_id

data = [
    {"text": "Today is hot, but I will manage!!!!", "label": 1},
    {"text": "Tomorrow is cold", "label": 0},
    {"text": "Carpe diem", "label": 1},
    {"text": "Tempus fugit", "label": 1},
]
dataset = Dataset.from_list(data)


def tokenize(row):
    return tokenizer(row["text"])


dataset = dataset.map(tokenize, remove_columns=["text"])
dataset.set_format(type="torch", columns=["input_ids", "attention_mask", "label"])
dataloader = data_utils.DataLoader(
    dataset, batch_size=100, collate_fn=DataCollatorWithPadding(tokenizer)
)

# %%
#
# Let's check the batch emitted by HF. We will see that it is a :code:`UserDict`,
# containing the input and label tensors. Note that :code:`UserDict` is
# :code:`MutableMapping`, so it is compatible with :code:`curvlinops`.

data = next(iter(dataloader))
print(f"Is the data a UserDict? {isinstance(data, UserDict)}")
for k, v in data.items():
    print(k, v.shape)


# %%
#
# Model
# -----
#
# Curvlinops supports general :code:`UserDict` inputs. However, everything must
# be handled inside the :code:`forward` function of the model. This gives
# the users the most flexibility, without much overhead.
#
# Let's wrap the HF model to conform this requirement then.


class MyGPT2(nn.Module):
    """
    Huggingface LLM wrapper.

    Args:
        tokenizer: The tokenizer used for preprocessing the text data. Needed
            since the model needs to know the padding token id.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer) -> None:
        super().__init__()
        config = GPT2Config.from_pretrained("gpt2")
        config.pad_token_id = tokenizer.pad_token_id
        config.num_labels = 2
        self.hf_model = GPT2ForSequenceClassification.from_pretrained(
            "gpt2", config=config
        )

        # For simplicity, only enable grad for the last layer
        for p in self.hf_model.parameters():
            p.requires_grad = False

        for p in self.hf_model.score.parameters():
            p.requires_grad = True

    def forward(self, data: MutableMapping) -> torch.Tensor:
        """
        Custom forward function. Handles things like moving the
        input tensor to the correct device inside.

        Args:
            data: A dict-like data structure with `input_ids` inside.
                This is the default data structure assumed by Huggingface
                dataloaders.

        Returns:
            logits: An `(batch_size, n_classes)`-sized tensor of logits.
        """
        device = next(self.parameters()).device
        input_ids = data["input_ids"].to(device)
        output_dict = self.hf_model(input_ids)
        return output_dict.logits


model = MyGPT2(tokenizer).to(torch.bfloat16)

with torch.no_grad():
    logits = model(data)
    print(f"Logits shape: {logits.shape}")


# %%
#
# Curvlinops
# ----------
#
# We are now ready to compute the curvature of this HF model using Curvlinops.
# For this, we need to define a function to tell Curvlinops how to get the
# batch size of the :code:`UserDict` input batch. Everything else is unchanged
# from the standard usage of Curvlinops!


def batch_size_fn(x: MutableMapping):
    return x["input_ids"].shape[0]


params = [p for p in model.parameters() if p.requires_grad]

ggn = GGNLinearOperator(
    model,
    nn.CrossEntropyLoss(),
    params,
    [(data, data["labels"])],  # We still need to input a list of "(X, y)" pairs!
    check_deterministic=False,
    batch_size_fn=batch_size_fn,  # Remember to specify this!
)

G = ggn @ np.eye(ggn.shape[0])

print(f"GGN shape: {G.shape}")


# %%
#
# Conclusion
# ----------
#
# This :code:`UserDict` (or any other dict-like data structure) specification
# is very flexible. This doesn't stop at HF models. You can leverage this
# for any custom models!



"""Implementation of Hutch++ trace estimation from Meyer et al."""

from typing import Optional, Union

from numpy import column_stack, dot, ndarray
from numpy.linalg import qr
from scipy.sparse.linalg import LinearOperator

from curvlinops.sampling import random_vector


class HutchPPTraceEstimator:
    r"""Class to perform trace estimation with the Huch++ method.

    In contrast to vanilla Hutchinson, Hutch++ has lower variance, but requires more
    memory.

    For details, see

    - Meyer, R. A., Musco, C., Musco, C., & Woodruff, D. P. (2020). Hutch++:
      optimal stochastic trace estimation.

    Let :math:`\mathbf{A}` be a square linear operator whose trace we want to
    approximate. First, we compute an orthonormal basis :math:`\mathbf{Q}` of a
    sub-space spanned by :math:`\mathbf{A} \mathbf{S}` where :math:`\mathbf{S}` is a
    tall random matrix with i.i.d. elements. Then, we compute the trace in the sub-space
    and apply Hutchinson's estimator in the remaining space spanned by
    :math:`\mathbf{I} - \mathbf{Q} \mathbf{Q}^\top`: We can draw a random vector
    :math:`\mathbf{v}` which satisfies
    :math:`\mathbb{E}[\mathbf{v} \mathbf{v}^\top] = \mathbf{I}` and sample from the
    estimator

    .. math::
        a
        := \mathrm{Tr}(\mathbf{Q}^\top \mathbf{A} \mathbf{Q})
        + \mathbf{v}^\top (\mathbf{I} - \mathbf{Q} \mathbf{Q}^\top)^\top
          \mathbf{A} (\mathbf{I} - \mathbf{Q} \mathbf{Q}^\top) \mathbf{v}
        \approx \mathrm{Tr}(\mathbf{A})\,.

    This estimator is unbiased, :math:`\mathbb{E}[a] = \mathrm{Tr}(\mathbf{A})`, as the
    first term is constant and the second part is Hutchinson's estimator in a sub-space.

    Example:
        >>> from numpy import trace, mean, round
        >>> from numpy.random import rand, seed
        >>> seed(0) # make deterministic
        >>> A = rand(10, 10)
        >>> tr_A = trace(A) # exact trace as reference
        >>> estimator = HutchPPTraceEstimator(A)
        >>> # one- and multi-sample approximations
        >>> tr_A_low_precision = estimator.sample()
        >>> tr_A_high_precision = mean([estimator.sample() for _ in range(998)])
        >>> # assert abs(tr_A - tr_A_low_precision) > abs(tr_A - tr_A_high_precision)
        >>> round(tr_A, 4), round(tr_A_low_precision, 4), round(tr_A_high_precision, 4)
        (4.4575, 2.4085, 4.5791)
    """

    def __init__(
        self,
        A: LinearOperator,
        basis_dim: Optional[int] = None,
        basis_distribution: str = "rademacher",
    ):
        """Store the linear operator whose trace will be estimated.

        Args:
            A: Linear square-shaped operator whose trace will be estimated.
            basis_dim: Dimension of the subspace used for exact trace estimation.
                Can be at most the linear operator's dimension. By default, its
                size will be 1% of the matrix's dimension, but at most ``10``.
                This assumes that we are working with very large matrices and we can
                only afford storing a small number of columns at a time.
            basis_distribution: Distribution of the vectors used to construct the
                subspace. Either ``'rademacher'` or ``'normal'``. Default is
                ``'rademacher'``.

        Raises:
            ValueError: If the operator is not square or the basis dimension is too
                large.

        Note:
            If you are planning to perform a fair (i.e. same computation budget)
            comparison with vanilla Hutchinson, ``basis_dim`` should be ``s / 3``
            where ``s`` is the number of samples used by vanilla Hutchinson. If
            ``s / 3`` requires storing a too large matrix, you can pick
            ``basis_dim = s1`` and draw ``s2`` samples from Hutch++ such that
            ``2 * s1 + s2 = s``.
        """
        if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
            raise ValueError(f"A must be square. Got shape {A.shape}.")
        self._A = A

        dim = A.shape[1]
        basis_dim = basis_dim if basis_dim is not None else min(10, max(dim // 100, 1))
        if basis_dim > self._A.shape[1]:
            raise ValueError(
                f"Basis dimension must be at most {self._A.shape[1]}. Got {basis_dim}."
            )
        self._basis_dim = basis_dim
        self._basis_distribution = basis_distribution

        # When drawing the first sample, the basis and its subspace trace will be
        # computed and stored in the following buffers for further samples
        self._Q: Union[ndarray, None] = None
        self._tr_QT_A_Q: Union[float, None] = None

    def sample(self, distribution: str = "rademacher") -> float:
        """Draw a sample from the trace estimator.

        Multiple samples can be combined into a more accurate trace estimation via
        averaging.

        Note:
            Calling this function for the first time will also compute the sub-space and
            its trace. Future calls will be faster as the latter are cached internally.

        Args:
            distribution: Distribution of the vector along which the linear operator
                will be evaluated. Either ``'rademacher'`` or ``'normal'``.
                Default is ``'rademacher'``.

        Returns:
            Sample from the trace estimator.
        """
        self.maybe_compute_and_cache_subspace()

        dim = self._A.shape[1]
        v = random_vector(dim, distribution)
        # project out subspace
        v -= self._Q @ (self._Q.T @ v)
        Av = self._A @ v
        return self._tr_QT_A_Q + dot(v, Av)

    def maybe_compute_and_cache_subspace(self):
        """Compute and cache the subspace and its trace if not already done."""
        if self._Q is not None and self._tr_QT_A_Q is not None:
            return

        dim = self._A.shape[1]
        AS = column_stack(
            [
                self._A @ random_vector(dim, self._basis_distribution)
                for _ in range(self._basis_dim)
            ]
        )
        self._Q, _ = qr(AS)

        self._tr_QT_A_Q = 0.0
        for i in range(self._basis_dim):
            v = self._Q[:, i]
            Av = self._A @ v
            self._tr_QT_A_Q += dot(v, Av)



"""Tests for ``curvlinops/_base``."""

from pytest import raises

from curvlinops._base import _LinearOperator


def test_check_deterministic(non_deterministic_case):
    """Test that non-deterministic behavior is recognized."""
    model_func, loss_func, params, data, batch_size_fn = non_deterministic_case

    with raises(RuntimeError):
        _LinearOperator(
            model_func,
            loss_func,
            params,
            data,
            batch_size_fn=batch_size_fn,
            check_deterministic=True,
        )



"""Vanilla Hutchinson trace estimation."""

from numpy import dot
from scipy.sparse.linalg import LinearOperator

from curvlinops.sampling import random_vector


class HutchinsonTraceEstimator:
    r"""Class to perform trace estimation with Hutchinson's method.

    For details, see

    - Hutchinson, M. (1989). A stochastic estimator of the trace of the influence
      matrix for laplacian smoothing splines. Communication in Statistics---Simulation
      and Computation.

    Let :math:`\mathbf{A}` be a square linear operator. We can approximate its trace
    :math:`\mathrm{Tr}(\mathbf{A})` by drawing a random vector :math:`\mathbf{v}`
    which satisfies :math:`\mathbb{E}[\mathbf{v} \mathbf{v}^\top] = \mathbf{I}` and
    sample from the estimator

    .. math::
        a
        := \mathbf{v}^\top \mathbf{A} \mathbf{v}
        \approx \mathrm{Tr}(\mathbf{A})\,.

    This estimator is unbiased,

    .. math::
        \mathbb{E}[a]
        = \mathrm{Tr}(\mathbb{E}[\mathbf{v}^\top\mathbf{A} \mathbf{v}])
        = \mathrm{Tr}(\mathbf{A} \mathbb{E}[\mathbf{v} \mathbf{v}^\top])
        = \mathrm{Tr}(\mathbf{A} \mathbf{I})
        = \mathrm{Tr}(\mathbf{A})\,.

    Example:
        >>> from numpy import trace, mean, round
        >>> from numpy.random import rand, seed
        >>> seed(0) # make deterministic
        >>> A = rand(10, 10)
        >>> tr_A = trace(A) # exact trace as reference
        >>> estimator = HutchinsonTraceEstimator(A)
        >>> # one- and multi-sample approximations
        >>> tr_A_low_precision = estimator.sample()
        >>> tr_A_high_precision = mean([estimator.sample() for _ in range(1_000)])
        >>> assert abs(tr_A - tr_A_low_precision) > abs(tr_A - tr_A_high_precision)
        >>> round(tr_A, 4), round(tr_A_low_precision, 4), round(tr_A_high_precision, 4)
        (4.4575, 6.6796, 4.3886)
    """

    def __init__(self, A: LinearOperator):
        """Store the linear operator whose trace will be estimated.

        Args:
            A: Linear square-shaped operator whose trace will be estimated.

        Raises:
            ValueError: If the operator is not square.
        """
        if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
            raise ValueError(f"A must be square. Got shape {A.shape}.")
        self._A = A

    def sample(self, distribution: str = "rademacher") -> float:
        """Draw a sample from the trace estimator.

        Multiple samples can be combined into a more accurate trace estimation via
        averaging.

        Args:
            distribution: Distribution of the vector along which the linear operator
                will be evaluated. Either ``'rademacher'`` or ``'normal'``.
                Default is ``'rademacher'``.

        Returns:
            Sample from the trace estimator.
        """
        dim = self._A.shape[1]
        v = random_vector(dim, distribution)
        Av = self._A @ v
        return dot(v, Av)



"""Contains pytest fixtures that are visible by other files."""

import test.utils
from collections.abc import MutableMapping
from test.cases import (
    ADJOINT_CASES,
    CASES,
    CNN_CASES,
    INV_CASES,
    NON_DETERMINISTIC_CASES,
)
from test.kfac_cases import (
    KFAC_EXACT_CASES,
    KFAC_EXACT_ONE_DATUM_CASES,
    KFAC_WEIGHT_SHARING_EXACT_CASES,
    SINGLE_LAYER_CASES,
    SINGLE_LAYER_WEIGHT_SHARING_CASES,
)
from typing import Callable, Dict, Iterable, List, Optional, Tuple

from numpy import random
from pytest import fixture
from torch import Tensor, manual_seed
from torch.nn import Module, MSELoss


def initialize_case(
    case: Dict,
) -> Tuple[
    Callable[[Tensor], Tensor],
    Callable[[Tensor, Tensor], Tensor],
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    random.seed(case["seed"])
    manual_seed(case["seed"])

    model_func = case["model_func"]().to(case["device"])
    loss_func = case["loss_func"]().to(case["device"])
    params = [p for p in model_func.parameters() if p.requires_grad]
    data = case["data"]()

    # In some KFAC cases,
    # ``data = {KFACType.EXPAND: [(X, y), ...], KFACType.REDUCE: [(X, y), ...]}``
    # unlike the standard ``data = [(X: Tensor | MutableMapping, y), ...]``.
    # We ignore the former since the latter is included in KFAC cases, and thus the
    # feature of ``MutableMapping`` inputs is sufficiently covered already.
    if not isinstance(data, dict) and isinstance(next(iter(data))[0], MutableMapping):
        batch_size_fn = test.utils.batch_size_fn
    else:
        batch_size_fn = None

    return model_func, loss_func, params, data, batch_size_fn


@fixture(params=CASES)
def case(
    request,
) -> Tuple[
    Callable[[Tensor], Tensor],
    Callable[[Tensor, Tensor], Tensor],
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    case = request.param
    yield initialize_case(case)


@fixture(params=INV_CASES)
def inv_case(
    request,
) -> Tuple[
    Callable[[Tensor], Tensor],
    Callable[[Tensor, Tensor], Tensor],
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    case = request.param
    yield initialize_case(case)


@fixture(params=CNN_CASES)
def cnn_case(
    request,
) -> Tuple[
    Callable[[Tensor], Tensor],
    Callable[[Tensor, Tensor], Tensor],
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    cnn_case = request.param
    yield initialize_case(cnn_case)


@fixture(params=NON_DETERMINISTIC_CASES)
def non_deterministic_case(
    request,
) -> Tuple[
    Callable[[Tensor], Tensor],
    Callable[[Tensor, Tensor], Tensor],
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    case = request.param
    yield initialize_case(case)


@fixture(params=ADJOINT_CASES)
def adjoint(request) -> bool:
    return request.param


@fixture(params=KFAC_EXACT_CASES)
def kfac_exact_case(
    request,
) -> Tuple[
    Module,
    MSELoss,
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    """Prepare a test case for which KFAC equals the GGN.

    Yields:
        A neural network, the mean-squared error function, a list of parameters, and
        a data set.
    """
    case = request.param
    yield initialize_case(case)


@fixture(params=KFAC_WEIGHT_SHARING_EXACT_CASES)
def kfac_weight_sharing_exact_case(
    request,
) -> Tuple[
    Module,
    MSELoss,
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    """Prepare a test case with weight-sharing for which KFAC equals the GGN.

    Yields:
        A neural network, the mean-squared error function, a list of parameters, and
        a data set.
    """
    case = request.param
    yield initialize_case(case)


@fixture(params=KFAC_EXACT_ONE_DATUM_CASES)
def kfac_exact_one_datum_case(
    request,
) -> Tuple[
    Module,
    Module,
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    """Prepare a test case for which KFAC equals the GGN and one datum is used.

    Yields:
        A neural network, loss function, a list of parameters, and
        a data set with a single datum.
    """
    case = request.param
    yield initialize_case(case)


@fixture(params=SINGLE_LAYER_CASES)
def single_layer_case(
    request,
) -> Tuple[
    Module,
    Module,
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    """Prepare a test case with a single-layer model for which FOOF is exact.

    Yields:
        A neural network, loss function, a list of parameters, and
        a data set with a single datum.
    """
    case = request.param
    yield initialize_case(case)


@fixture(params=SINGLE_LAYER_WEIGHT_SHARING_CASES)
def single_layer_weight_sharing_case(
    request,
) -> Tuple[
    Module,
    Module,
    List[Tensor],
    Iterable[Tuple[Tensor, Tensor]],
    Optional[Callable[[MutableMapping], int]],
]:
    """Test case with a single-layer model with weight-sharing for which FOOF is exact.

    Yields:
        A neural network, loss function, a list of parameters, and
        a data set with a single datum.
    """
    case = request.param
    yield initialize_case(case)



r"""
Matrix-vector products
======================

This tutorial contains a basic demonstration how to set up ``LinearOperators``
for the Hessian and the GGN and how to multiply them to a vector.

First, the imports.
"""

import matplotlib.pyplot as plt
import numpy
import torch
from torch import nn

from curvlinops import GGNLinearOperator, HessianLinearOperator
from curvlinops.examples.functorch import functorch_ggn, functorch_hessian
from curvlinops.examples.utils import report_nonclose

# make deterministic
torch.manual_seed(0)
numpy.random.seed(0)

# %%
# Setup
# -----
# Let's create some toy data, a small MLP, and use mean-squared error as loss function.

N = 4
D_in = 7
D_hidden = 5
D_out = 3

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

X = torch.rand(N, D_in).to(DEVICE)
y = torch.rand(N, D_out).to(DEVICE)

model = nn.Sequential(
    nn.Linear(D_in, D_hidden),
    nn.ReLU(),
    nn.Linear(D_hidden, D_hidden),
    nn.Sigmoid(),
    nn.Linear(D_hidden, D_out),
).to(DEVICE)
params = [p for p in model.parameters() if p.requires_grad]

loss_function = nn.MSELoss(reduction="mean").to(DEVICE)


# %%
# Hessian-vector products
# -----------------------
#
# Setting up a linear operator for the Hessian is straightforward.

data = [(X, y)]
H = HessianLinearOperator(model, loss_function, params, data)

# %%
#
# We can now multiply by the Hessian. This operation will be carried out in
# PyTorch under the hood, but the operator is compatible with ``scipy``, so we
# can just pass a ``numpy`` vector to the matrix-multiplication.

D = H.shape[0]
v = numpy.random.rand(D)

Hv = H @ v

# %%
#
# To verify the result, we compute the Hessian using ``functorch``, using a
# utility function from ``curvlinops.examples``:

H_mat = functorch_hessian(model, loss_function, params, data).detach().cpu().numpy()

# %%
#
# Let's check that the multiplication onto ``v`` leads to the same result:

Hv_functorch = H_mat @ v

print("Comparing Hessian-vector product with functorch's Hessian-vector product.")
report_nonclose(Hv, Hv_functorch)


# %%
# Hessian-matrix products
# -----------------------
#
# We can also compute the Hessian's matrix representation with the linear
# operator, simply by multiplying it onto the identity matrix. (Of course, this
# only works if the Hessian is small enough.)
H_mat_from_linop = H @ numpy.eye(D)

# %%
#
# This should yield the same matrix as with :code:`functorch`.

print("Comparing Hessian with functorch's Hessian.")
report_nonclose(H_mat, H_mat_from_linop)

# %%
#
# Last, here's a visualization of the Hessian.

plt.figure()
plt.title("Hessian")
plt.imshow(H_mat)
plt.colorbar()

# %%
# GGN-vector products
# -------------------
#
# Setting up a linear operator for the Fisher/GGN is identical to the Hessian.

GGN = GGNLinearOperator(model, loss_function, params, data)

# %%
#
# Let's compute a GGN-vector product.

D = H.shape[0]
v = numpy.random.rand(D)

GGNv = GGN @ v

# %%
#
# To verify the result, we will use ``functorch`` to compute the GGN. For that,
# we use that the GGN corresponds to the Hessian if we replace the neural
# network by its linearization. This is implemented in a utility function of
# :code:`curvlinops.examples`:

GGN_mat = functorch_ggn(model, loss_function, params, data).detach().cpu().numpy()

GGNv_functorch = GGN_mat @ v

print("Comparing GGN-vector product with functorch's GGN-vector product.")
report_nonclose(GGNv, GGNv_functorch)

# %%
# GGN-matrix products
# -------------------
#
# We can also compute the GGN matrix representation with the linear operator,
# simply by multiplying it onto the identity matrix. (Of course, this only
# works if the GGN is small enough.)
GGN_mat_from_linop = GGN @ numpy.eye(D)

# %%
#
# This should yield the same matrix as with :code:`functorch`.

print("Comparing GGN with functorch's GGN.")
report_nonclose(GGN_mat, GGN_mat_from_linop)

# %%
#
# Last, here's a visualization of the GGN.

plt.figure()
plt.title("GGN")
plt.imshow(GGN_mat)
plt.colorbar()

# %%
# Visual comparison: Hessian and GGN
# ----------------------------------
#
# To conclude, let's plot both the Hessian and GGN using the same limits

min_value = min(GGN_mat.min(), H_mat.min())
max_value = max(GGN_mat.max(), H_mat.max())

fig, ax = plt.subplots(ncols=2)
ax[0].set_title("Hessian")
ax[0].imshow(H_mat, vmin=min_value, vmax=max_value)
ax[1].set_title("GGN")
ax[1].imshow(GGN_mat, vmin=min_value, vmax=max_value)



"""Contains test cases for the KFAC linear operator."""

from collections import UserDict
from functools import partial
from test.cases import ModelWithDictInput
from test.utils import (
    Conv2dModel,
    WeightShareModel,
    binary_classification_targets,
    classification_targets,
    get_available_devices,
    regression_targets,
)

from torch import rand
from torch.nn import (
    BCEWithLogitsLoss,
    CrossEntropyLoss,
    Identity,
    Linear,
    MSELoss,
    Sequential,
)

from curvlinops.kfac import KFACType

# Add test cases here, devices and loss function with different reductions will be
# added automatically below
KFAC_EXACT_CASES_NO_DEVICE_NO_LOSS_FUNC = [
    ###############################################################################
    #                                  REGRESSION                                 #
    ###############################################################################
    # deep linear network with scalar output
    {
        "model_func": lambda: Sequential(Linear(6, 3), Linear(3, 1)),
        "data": lambda: [
            (rand(2, 6), regression_targets((2, 1))),
            (rand(5, 6), regression_targets((5, 1))),
        ],
        "seed": 0,
    },
    # deep linear network with vector output
    {
        "model_func": lambda: Sequential(Linear(5, 4), Linear(4, 3)),
        "data": lambda: [
            (rand(1, 5), regression_targets((1, 3))),
            (rand(7, 5), regression_targets((7, 3))),
        ],
        "seed": 0,
    },
    # deep linear network with dict-like input
    {
        "model_func": lambda: ModelWithDictInput(num_classes=3, nonlin=Identity),
        "data": lambda: [
            (UserDict({"x": rand(1, 10)}), regression_targets((1, 3))),
            ({"x": rand(7, 10)}, regression_targets((7, 3))),
        ],
        "seed": 0,
    },
]

KFAC_EXACT_CASES = []
for case in KFAC_EXACT_CASES_NO_DEVICE_NO_LOSS_FUNC:
    for device in get_available_devices():
        for reduction in ["mean", "sum"]:
            case_with_device_and_loss_func = {
                **case,
                "device": device,
                "loss_func": partial(MSELoss, reduction=reduction),
            }
            KFAC_EXACT_CASES.append(case_with_device_and_loss_func)


# Add test cases here, devices and loss function with different reductions will be
# added automatically below
KFAC_WEIGHT_SHARING_EXACT_CASES_NO_DEVICE_NO_LOSS_FUNC = [
    ###############################################################################
    #                                  REGRESSION                                 #
    ###############################################################################
    # deep linear network with vector output and weight-sharing dimensions
    {
        "model_func": lambda: WeightShareModel(Linear(5, 4), Linear(4, 3)),
        "data": lambda: {
            KFACType.EXPAND: [
                (rand(2, 4, 8, 5), regression_targets((2, 4, 8, 3))),
                (rand(7, 4, 8, 5), regression_targets((7, 4, 8, 3))),
            ],
            KFACType.REDUCE: [
                (rand(1, 4, 8, 5), regression_targets((1, 3))),
                (rand(7, 4, 8, 5), regression_targets((7, 3))),
            ],
        },
        "seed": 0,
    },
    # Conv2d module with vector output (uses average pooling for reduce setting)
    {
        "model_func": lambda: Conv2dModel(),
        "data": lambda: {
            KFACType.EXPAND: [
                (rand(2, 3, 32, 32), regression_targets((2, 33, 33, 2))),
                (rand(7, 3, 32, 32), regression_targets((7, 33, 33, 2))),
            ],
            KFACType.REDUCE: [
                (rand(1, 3, 32, 32), regression_targets((1, 2))),
                (rand(8, 3, 32, 32), regression_targets((8, 2))),
            ],
        },
        "seed": 0,
    },
]

KFAC_WEIGHT_SHARING_EXACT_CASES = []
for case in KFAC_WEIGHT_SHARING_EXACT_CASES_NO_DEVICE_NO_LOSS_FUNC:
    for device in get_available_devices():
        for reduction in ["mean", "sum"]:
            case_with_device_and_loss_func = {
                **case,
                "device": device,
                "loss_func": partial(MSELoss, reduction=reduction),
            }
            KFAC_WEIGHT_SHARING_EXACT_CASES.append(case_with_device_and_loss_func)


# Add test cases here, devices will be added automatically below
KFAC_EXACT_ONE_DATUM_CASES_NO_DEVICE = [
    ###############################################################################
    #                                CLASSIFICATION                               #
    ###############################################################################
    # deep linear network with vector output (both reductions)
    {
        "model_func": lambda: Sequential(Linear(5, 4), Linear(4, 3)),
        "loss_func": lambda: CrossEntropyLoss(reduction="mean"),
        "data": lambda: [(rand(1, 5), classification_targets((1,), 3))],
        "seed": 0,
    },
    {
        "model_func": lambda: Sequential(Linear(5, 4), Linear(4, 3)),
        "loss_func": lambda: CrossEntropyLoss(reduction="sum"),
        "data": lambda: [(rand(1, 5), classification_targets((1,), 3))],
        "seed": 0,
    },
    # deep linear network with vector output and BCEWithLogitsLoss (both reductions)
    {
        "model_func": lambda: Sequential(Linear(4, 3), Linear(3, 2)),
        "loss_func": lambda: BCEWithLogitsLoss(reduction="mean"),
        "data": lambda: [(rand(1, 4), binary_classification_targets((1, 2)))],
        "seed": 0,
    },
    {
        "model_func": lambda: Sequential(Linear(4, 3), Linear(3, 2)),
        "loss_func": lambda: BCEWithLogitsLoss(reduction="sum"),
        "data": lambda: [(rand(1, 4), binary_classification_targets((1, 2)))],
        "seed": 0,
    },
]

KFAC_EXACT_ONE_DATUM_CASES = []
for case in KFAC_EXACT_ONE_DATUM_CASES_NO_DEVICE:
    for device in get_available_devices():
        case_with_device = {
            **case,
            "device": device,
        }
        KFAC_EXACT_ONE_DATUM_CASES.append(case_with_device)


# Add test cases here, devices and loss function with different reductions will be
# added automatically below
SINGLE_LAYER_CASES_NO_DEVICE_NO_LOSS_FUNC = [
    ###############################################################################
    #                                  REGRESSION                                 #
    ###############################################################################
    # single-layer linear network with scalar output
    {
        "model_func": lambda: Linear(6, 1),
        "data": lambda: [
            (rand(5, 6), regression_targets((5, 1))),
            (rand(5, 6), regression_targets((5, 1))),
        ],
        "seed": 0,
    },
    # single-layer linear network with vector output
    {
        "model_func": lambda: Linear(5, 3),
        "data": lambda: [
            (rand(7, 5), regression_targets((7, 3))),
            (rand(7, 5), regression_targets((7, 3))),
        ],
        "seed": 0,
    },
]

SINGLE_LAYER_CASES = []
for case in SINGLE_LAYER_CASES_NO_DEVICE_NO_LOSS_FUNC:
    for device in get_available_devices():
        for reduction in ["mean", "sum"]:
            case_with_device_and_loss_func = {
                **case,
                "device": device,
                "loss_func": partial(MSELoss, reduction=reduction),
            }
            SINGLE_LAYER_CASES.append(case_with_device_and_loss_func)


# Add test cases here, devices and loss function with different reductions will be
# added automatically below
SINGLE_LAYER_WEIGHT_SHARING_CASES_NO_DEVICE_NO_LOSS_FUNC = [
    ###############################################################################
    #                                  REGRESSION                                 #
    ###############################################################################
    # single linear layer with vector output and weight-sharing dimensions
    {
        "model_func": lambda: WeightShareModel(Linear(5, 3)),
        "data": lambda: {
            KFACType.EXPAND: [
                (rand(7, 4, 8, 5), regression_targets((7, 4, 8, 3))),
                (rand(7, 4, 8, 5), regression_targets((7, 4, 8, 3))),
            ],
            KFACType.REDUCE: [
                (rand(8, 4, 8, 5), regression_targets((8, 3))),
                (rand(8, 4, 8, 5), regression_targets((8, 3))),
            ],
        },
        "seed": 0,
    },
    # Conv2d module with vector output (uses average pooling for reduce setting)
    {
        "model_func": lambda: Conv2dModel(),
        "data": lambda: {
            KFACType.EXPAND: [
                (rand(7, 3, 32, 32), regression_targets((7, 33, 33, 2))),
                (rand(7, 3, 32, 32), regression_targets((7, 33, 33, 2))),
            ],
            KFACType.REDUCE: [
                (rand(8, 3, 32, 32), regression_targets((8, 2))),
                (rand(8, 3, 32, 32), regression_targets((8, 2))),
            ],
        },
        "seed": 0,
    },
]

SINGLE_LAYER_WEIGHT_SHARING_CASES = []
for case in SINGLE_LAYER_WEIGHT_SHARING_CASES_NO_DEVICE_NO_LOSS_FUNC:
    for device in get_available_devices():
        for reduction in ["mean", "sum"]:
            case_with_device_and_loss_func = {
                **case,
                "device": device,
                "loss_func": partial(MSELoss, reduction=reduction),
            }
            SINGLE_LAYER_WEIGHT_SHARING_CASES.append(case_with_device_and_loss_func)



"""Sampling methods for random vectors."""

from numpy import ndarray
from numpy.random import binomial, randn


def rademacher(dim: int) -> ndarray:
    """Draw a vector with i.i.d. Rademacher elements.

    Args:
        dim: Dimension of the vector.

    Returns:
        Vector with i.i.d. Rademacher elements and specified dimension.
    """
    num_trials, success_prob = 1, 0.5
    return binomial(num_trials, success_prob, size=dim).astype(float) * 2 - 1


def normal(dim: int) -> ndarray:
    """Draw a vector with i.i.d. standard normal elements.

    Args:
        dim: Dimension of the vector.

    Returns:
        Vector with i.i.d. standard normal elements and specified dimension.
    """
    return randn(dim)


def random_vector(dim: int, distribution: str) -> ndarray:
    """Draw a vector with i.i.d. elements.

    Args:
        dim: Dimension of the vector.
        distribution: Distribution of the vector's elements. Either ``'rademacher'`` or
            ``'normal'``.

    Returns:
        Vector with i.i.d. elements and specified dimension.

    Raises:
        ValueError: If the distribution is unknown.
    """
    if distribution == "rademacher":
        return rademacher(dim)
    elif distribution == "normal":
        return normal(dim)
    else:
        raise ValueError(f"Unknown distribution {distribution:!r}.")



"""``curvlinops`` library API."""

from curvlinops.diagonal.hutchinson import HutchinsonDiagonalEstimator
from curvlinops.fisher import FisherMCLinearOperator
from curvlinops.ggn import GGNLinearOperator
from curvlinops.gradient_moments import EFLinearOperator
from curvlinops.hessian import HessianLinearOperator
from curvlinops.inverse import (
    CGInverseLinearOperator,
    KFACInverseLinearOperator,
    LSMRInverseLinearOperator,
    NeumannInverseLinearOperator,
)
from curvlinops.jacobian import JacobianLinearOperator, TransposedJacobianLinearOperator
from curvlinops.kfac import FisherType, KFACLinearOperator, KFACType
from curvlinops.norm.hutchinson import HutchinsonSquaredFrobeniusNormEstimator
from curvlinops.papyan2020traces.spectrum import (
    LanczosApproximateLogSpectrumCached,
    LanczosApproximateSpectrumCached,
    lanczos_approximate_log_spectrum,
    lanczos_approximate_spectrum,
)
from curvlinops.submatrix import SubmatrixLinearOperator
from curvlinops.trace.hutchinson import HutchinsonTraceEstimator
from curvlinops.trace.meyer2020hutch import HutchPPTraceEstimator

__all__ = [
    # linear operators
    "HessianLinearOperator",
    "GGNLinearOperator",
    "EFLinearOperator",
    "FisherMCLinearOperator",
    "KFACLinearOperator",
    "JacobianLinearOperator",
    "TransposedJacobianLinearOperator",
    # Enums
    "FisherType",
    "KFACType",
    # inversion
    "CGInverseLinearOperator",
    "LSMRInverseLinearOperator",
    "NeumannInverseLinearOperator",
    "KFACInverseLinearOperator",
    # slicing
    "SubmatrixLinearOperator",
    # spectral properties
    "lanczos_approximate_spectrum",
    "lanczos_approximate_log_spectrum",
    "LanczosApproximateSpectrumCached",
    "LanczosApproximateLogSpectrumCached",
    # trace estimation
    "HutchinsonTraceEstimator",
    "HutchPPTraceEstimator",
    # diagonal estimation
    "HutchinsonDiagonalEstimator",
    # norm estimation
    "HutchinsonSquaredFrobeniusNormEstimator",
]
